from aana.configs.deployments import (
    hf_blip2_opt_2_7b_deployment,
    internlm2_5_7b_chat_deployment,
    vad_deployment,
    whisper_medium_deployment,
)
from aana.projects.chat_with_video.endpoints import (
    DeleteMediaEndpoint,
    GetVideoStatusEndpoint,
    IndexVideoEndpoint,
    LoadVideoMetadataEndpoint,
    VideoChatEndpoint,
)
from aana.sdk import AanaSDK

deployments = [
    {
        "name": "asr_deployment",
        "instance": whisper_medium_deployment,
    },
    {
        "name": "vad_deployment",
        "instance": vad_deployment,
    },
    {
        "name": "captioning_deployment",
        "instance": hf_blip2_opt_2_7b_deployment,
    },
    {
        "name": "llm_deployment",
        "instance": internlm2_5_7b_chat_deployment,
    },
]

endpoints = [
    {
        "name": "index_video_stream",
        "path": "/video/index_stream",
        "summary": "Index a video and return the captions and transcriptions as a stream",
        "endpoint_cls": IndexVideoEndpoint,
    },
    {
        "name": "video_metadata",
        "path": "/video/metadata",
        "summary": "Load video metadata",
        "endpoint_cls": LoadVideoMetadataEndpoint,
    },
    {
        "name": "video_chat_stream",
        "path": "/video/chat_stream",
        "summary": "Chat with video using LLaMa2 7B Chat (streaming)",
        "endpoint_cls": VideoChatEndpoint,
    },
    {
        "name": "video_status",
        "path": "/video/status",
        "summary": "Get video status",
        "endpoint_cls": GetVideoStatusEndpoint,
    },
    {
        "name": "delete_media",
        "path": "/video/delete",
        "summary": "Delete a media",
        "endpoint_cls": DeleteMediaEndpoint,
    },
]

aana_app = AanaSDK(name="chat_with_video_app")

for deployment in deployments:
    aana_app.register_deployment(
        name=deployment["name"],
        instance=deployment["instance"],
    )

for endpoint in endpoints:
    aana_app.register_endpoint(
        name=endpoint["name"],
        path=endpoint["path"],
        summary=endpoint["summary"],
        endpoint_cls=endpoint["endpoint_cls"],
    )



from enum import Enum

from sqlalchemy import Column, Float, ForeignKey, String
from sqlalchemy import Enum as SqlEnum
from sqlalchemy.orm import relationship

from aana.storage.models.base import BaseEntity, TimeStampEntity
from aana.storage.types import MediaIdSqlType


class Status(str, Enum):
    """Enum for video status."""

    CREATED = "created"
    RUNNING = "running"
    COMPLETED = "completed"
    FAILED = "failed"


class VideoEntity(BaseEntity, TimeStampEntity):
    """ORM class for video file (video, etc)."""

    __tablename__ = "video"

    id = Column(MediaIdSqlType, ForeignKey("media.id"), primary_key=True)
    duration = Column(Float, comment="Media duration in seconds")
    orig_filename = Column(String, comment="Original filename")
    orig_url = Column(String, comment="Original URL")
    title = Column(String, comment="Title of the video")
    description = Column(String, comment="Description of the video")
    status = Column(
        SqlEnum(Status),
        nullable=False,
        default=Status.CREATED,
        comment="Status of the video",
    )
    media = relationship("MediaEntity", back_populates="video", uselist=False)

    captions = relationship(
        "CaptionEntity",
        back_populates="video",
        cascade="all, delete",
        uselist=True,
    )
    transcripts = relationship(
        "TranscriptEntity",
        back_populates="video",
        cascade="all, delete",
        uselist=True,
    )



# ruff: noqa: A002
from pathlib import Path
from typing import TypedDict
from urllib.parse import urlparse

from sqlalchemy.orm import Session

from aana.core.models.asr import (
    AsrSegment,
    AsrSegments,
    AsrTranscription,
    AsrTranscriptionInfo,
)
from aana.core.models.captions import CaptionsList
from aana.core.models.media import MediaId
from aana.core.models.video import Video, VideoMetadata
from aana.storage.engine import engine
from aana.storage.models import (
    CaptionEntity,
    MediaEntity,
    MediaType,
    TranscriptEntity,
    VideoEntity,
)
from aana.storage.models.video import Status as VideoStatus
from aana.storage.repository.caption import CaptionRepository
from aana.storage.repository.media import MediaRepository
from aana.storage.repository.transcript import TranscriptRepository
from aana.storage.repository.video import VideoRepository


# Just using raw utility functions like this isn't a permanent solution, but
# it's good enough for now to validate what we're working on.
def save_video_batch(
    videos: list[Video],
    durations: list[float],
    # , video_params: list[VideoParams]
) -> dict:
    """Saves a batch of videos to datastore.

    Args:
        videos (list[Video]): The video objects.
        durations (list[float]): the duration of each video object

    Returns:
        dict: The dictionary with video and media IDs.
    """
    media_entities = []
    video_entities = []
    for video_object, duration in zip(videos, durations, strict=True):
        if video_object.url is not None:
            orig_url = video_object.url
            parsed_url = urlparse(orig_url)
            orig_filename = Path(parsed_url.path).name
        elif video_object.path is not None:
            parsed_path = Path(video_object.path)
            orig_filename = parsed_path.name
            orig_url = None
        else:
            orig_url = None
            orig_filename = None
        media_entity = MediaEntity(
            id=video_object.media_id,
            media_type=MediaType.VIDEO,
        )
        video_entity = VideoEntity(
            media=media_entity,
            duration=duration,
            orig_filename=orig_filename,
            orig_url=orig_url,
        )
        media_entities.append(media_entity)
        video_entities.append(video_entity)
    with Session(engine) as session:
        m_repo = MediaRepository(session)
        v_repo = VideoRepository(session)
        media_entities = m_repo.create_multiple(media_entities)
        video_entities = v_repo.create_multiple(video_entities)
        return {"media_ids": [m.id for m in media_entities]}


def save_video(video: Video, duration: float) -> dict:
    """Saves a video to datastore.

    Args:
        video (Video): The video object.
        duration (float): the duration of the video object

    Returns:
        dict: The dictionary with video and media IDs.
    """
    if video.url is not None:
        orig_url = video.url
        parsed_url = urlparse(orig_url)
        orig_filename = Path(parsed_url.path).name
    elif video.path is not None:
        parsed_path = Path(video.path)
        orig_filename = parsed_path.name
        orig_url = None
    else:
        orig_url = None
        orig_filename = None
    media_entity = MediaEntity(id=video.media_id, media_type=MediaType.VIDEO)
    video_entity = VideoEntity(
        id=video.media_id,
        orig_filename=orig_filename,
        orig_url=orig_url,
        title=video.title,
        duration=duration,
        description=video.description,
    )
    with Session(engine) as session:
        media_repo = MediaRepository(session)
        _ = media_repo.create(media_entity)
        video_repo = VideoRepository(session)
        _ = video_repo.create(video_entity)
        return {
            "media_id": media_entity.id,
        }


def get_video_status(
    media_id: MediaId,
) -> VideoStatus:
    """Load video status from database.

    Args:
        media_id (MediaId): The media ID.

    Returns:
        VideoStatus: The video status.
    """
    with Session(engine) as session:
        entity: VideoEntity = VideoRepository(session).read(media_id)
        return entity.status


def check_media_id_exist(
    media_id: MediaId,
) -> bool:
    """Check if media_id exists in the database.

    Args:
        media_id (MediaId): The media ID.

    Returns:
        bool: True if the media_id exists, False otherwise.
    """
    with Session(engine) as session:
        return MediaRepository(session).check_media_exists(media_id)


def update_video_status(media_id: MediaId, status: VideoStatus):
    """Update the video status.

    Args:
        media_id (str): The media ID.
        status (VideoStatus): The new status.
    """
    with Session(engine) as session:
        video_repo = VideoRepository(session)
        media_entity = video_repo.read(media_id)
        media_entity.status = status
        session.commit()


def delete_media(media_id: MediaId) -> dict:
    """Deletes a media file from the database.

    Args:
        media_id (MediaId): The media ID.

    Returns:
        dict: The dictionary with the ID of the deleted entity.
    """
    with Session(engine) as session:
        media_repo = MediaRepository(session)
        media_entity = media_repo.delete(media_id, check=True)
    return {"media_id": media_entity.id}


class SaveVideoCaptionOutput(TypedDict):
    """The output of the save video caption endpoint."""

    caption_ids: list[int]


def save_video_captions(
    model_name: str,
    media_id: MediaId,
    captions: CaptionsList,
    timestamps: list[float],
    frame_ids: list[int],
) -> SaveVideoCaptionOutput:
    """Save captions.

    Args:
        model_name (str): The name of the model used to generate the captions.
        media_id (MediaId): the media ID of the video.
        captions (CaptionsList): The captions.
        timestamps (list[float]): The timestamps.
        frame_ids (list[int]): The frame IDs.

    Returns:
        SaveVideoCaptionOutput: The dictionary with the IDs of the saved entities.
    """
    with Session(engine) as session:
        entities = [
            CaptionEntity.from_caption_output(
                model_name, media_id, frame_id, timestamp, caption
            )
            for caption, timestamp, frame_id in zip(
                captions, timestamps, frame_ids, strict=True
            )
        ]
        repo = CaptionRepository(session)
        results = repo.create_multiple(entities)
        return {
            "caption_ids": [c.id for c in results]  # type: ignore
        }


def save_video_transcription(
    model_name: str,
    media_id: MediaId,
    transcription_info: AsrTranscriptionInfo,
    transcription: AsrTranscription,
    segments: AsrSegments,
) -> dict:
    """Save transcripts.

    Args:
        model_name (str): The name of the model used to generate the transcript.
        media_id (MediaId): The media id of the video
        transcription_info (AsrTranscriptionInfo): The ASR transcription info.
        transcription (AsrTranscription): The ASR transcription.
        segments (AsrSegments): The ASR segments.

    Returns:
        dict: The dictionary with the IDs of the saved entities.
    """
    with Session(engine) as session:
        entities = [
            TranscriptEntity.from_asr_output(
                model_name=model_name,
                media_id=media_id,
                transcription=transcription,
                segments=segments,
                info=transcription_info,
            )
        ]

        repo = TranscriptRepository(session)
        entities = repo.create_multiple(entities)
        transcription_id = entities[0].id

        return {
            "transcription_id": transcription_id,
        }


def save_transcripts_batch(
    model_name: str,
    media_ids: list[MediaId],
    transcription_info_list: list[AsrTranscriptionInfo],
    transcription_list: list[AsrTranscription],
    segments_list: list[list[AsrSegment]],
) -> dict:
    """Save transcripts in a batch.

    Arguments:
        model_name (str): the nameof the model used to generate the batch of transcripts
        media_ids (list[MediaId]): the media ids of each video in the batch
        transcription_info_list (list[AsrTranscriptionInfo]): list of transcript metadata
        transcription_list (list[AsrTranscription]): list of transcripts
        segments_list (list[list[AsrSegment]]): list of audio segment definitions

    Returns:
        dict with key "transcription_ids"
    """
    with Session(engine) as session:
        entities = [
            TranscriptEntity.from_asr_output(
                model_name,
                media_id,
                transcript_info,
                transcript,
                segments,
            )
            for media_id, transcript_info, transcript, segments in zip(
                media_ids,
                transcription_info_list,
                transcription_list,
                segments_list,
                strict=True,
            )
        ]
        repo = TranscriptRepository(session)
        entities = repo.create_multiple(entities)
        return {
            "transcription_ids": [c.id for c in entities]  # type: ignore
        }


class LoadTranscriptionOutput(TypedDict):
    """The output of the load transcription endpoint."""

    transcription: AsrTranscription
    segments: AsrSegments
    transcription_info: AsrTranscriptionInfo


def load_video_transcription(
    model_name: str,
    media_id: MediaId,
) -> LoadTranscriptionOutput:
    """Load transcript from database.

    Args:
        model_name (str): The name of the model used to generate the transcript.
        media_id (MediaId): The media ID.

    Returns:
        LoadTranscriptionOutput: The dictionary with the transcript, segments, and info.
    """
    with Session(engine) as session:
        repo = TranscriptRepository(session)
        entity: TranscriptEntity = repo.get_transcript(model_name, media_id)
        transcription = AsrTranscription(text=entity.transcript)
        segments = [AsrSegment(**s) for s in entity.segments]
        info = AsrTranscriptionInfo(
            language=entity.language,
            language_confidence=entity.language_confidence,
        )
        return {
            "transcription": transcription,
            "segments": segments,
            "transcription_info": info,
        }


class LoadVideoCaptionsOutput(TypedDict):
    """The output of the load video captions endpoint."""

    captions: CaptionsList
    timestamps: list[float]
    frame_ids: list[int]


def load_video_captions(
    model_name: str,
    media_id: MediaId,
) -> LoadVideoCaptionsOutput:
    """Load captions from database.

    Args:
        model_name (str): The name of the model used to generate the captions.
        media_id (MediaId): The media ID.

    Returns:
        LoadVideoCaptionsOutput: The dictionary with the captions, timestamps, and frame IDs.
            captions: CaptionsList
            timestamps: list[float]
            frame_ids: list[int]
    """
    with Session(engine) as session:
        repo = CaptionRepository(session)
        entities = repo.get_captions(model_name, media_id)
        captions = [c.caption for c in entities]
        timestamps = [c.timestamp for c in entities]
        frame_ids = [c.frame_id for c in entities]
        return {
            "captions": captions,
            "timestamps": timestamps,
            "frame_ids": frame_ids,
        }


def load_video_metadata(
    media_id: MediaId,
) -> VideoMetadata:
    """Load video metadata from database.

    Args:
        media_id (MediaId): The media ID.

    Returns:
        VideoMetadata: The video metadata.
    """
    with Session(engine) as session:
        video_entity = VideoRepository(session).read(media_id)
        return VideoMetadata(
            title=video_entity.title,
            description=video_entity.description,
            duration=video_entity.duration,
        )



from collections.abc import AsyncGenerator
from typing import Annotated, TypedDict

from pydantic import Field
from transformers import BitsAndBytesConfig

from aana.api.api_generation import Endpoint
from aana.core.models.chat import ChatDialog, ChatMessage
from aana.core.models.video import VideoInput
from aana.deployments.aana_deployment_handle import AanaDeploymentHandle
from aana.deployments.hf_text_generation_deployment import (
    HfTextGenerationConfig,
    HfTextGenerationDeployment,
)
from aana.deployments.whisper_deployment import (
    WhisperComputeType,
    WhisperConfig,
    WhisperDeployment,
    WhisperModelSize,
    WhisperOutput,
)
from aana.integrations.external.yt_dlp import download_video
from aana.processors.remote import run_remote
from aana.processors.video import extract_audio
from aana.sdk import AanaSDK

# Define the model deployments.
asr_deployment = WhisperDeployment.options(
    num_replicas=1,
    ray_actor_options={"num_gpus": 0.25},
    user_config=WhisperConfig(
        model_size=WhisperModelSize.MEDIUM,
        compute_type=WhisperComputeType.FLOAT16,
    ).model_dump(mode="json"),
)

llm_deployment = HfTextGenerationDeployment.options(
    num_replicas=1,
    ray_actor_options={"num_gpus": 0.25},
    user_config=HfTextGenerationConfig(
        model_id="microsoft/Phi-3-mini-4k-instruct",
        model_kwargs={
            "trust_remote_code": True,
            "quantization_config": BitsAndBytesConfig(
                load_in_8bit=False, load_in_4bit=True
            ),
        },
    ).model_dump(mode="json"),
)


deployments = [
    {"name": "asr_deployment", "instance": asr_deployment},
    {"name": "llm_deployment", "instance": llm_deployment},
]


class SummarizeVideoEndpointOutput(TypedDict):
    """Summarize video endpoint output."""

    summary: Annotated[str, Field(description="The summary of the video.")]


class SummarizeVideoStreamEndpointOutput(TypedDict):
    """Summarize video endpoint output."""

    text: Annotated[str, Field(description="The text chunk.")]


# Define the endpoint to transcribe the video.
class TranscribeVideoEndpoint(Endpoint):
    """Transcribe video endpoint."""

    async def initialize(self):
        """Initialize the endpoint."""
        self.asr_handle = await AanaDeploymentHandle.create("asr_deployment")
        await super().initialize()

    async def run(self, video: VideoInput) -> WhisperOutput:
        """Transcribe video."""
        video_obj = await run_remote(download_video)(video_input=video)
        audio = extract_audio(video=video_obj)
        transcription = await self.asr_handle.transcribe(audio=audio)
        return transcription


class SummarizeVideoEndpoint(Endpoint):
    """Summarize video endpoint."""

    async def initialize(self):
        """Initialize the endpoint."""
        self.asr_handle = await AanaDeploymentHandle.create("asr_deployment")
        self.llm_handle = await AanaDeploymentHandle.create("llm_deployment")
        await super().initialize()

    async def run(self, video: VideoInput) -> SummarizeVideoEndpointOutput:
        """Summarize video."""
        video_obj = await run_remote(download_video)(video_input=video)
        audio = extract_audio(video=video_obj)
        transcription = await self.asr_handle.transcribe(audio=audio)
        transcription_text = transcription["transcription"].text
        dialog = ChatDialog(
            messages=[
                ChatMessage(
                    role="system",
                    content="You are a helpful assistant that can summarize audio transcripts.",
                ),
                ChatMessage(
                    role="user",
                    content=f"Summarize the following video transcript into a list of bullet points: {transcription_text}",
                ),
            ]
        )
        summary_response = await self.llm_handle.chat(dialog=dialog)
        summary_message: ChatMessage = summary_response["message"]
        summary = summary_message.content
        return {"summary": summary}


class SummarizeVideoStreamEndpoint(Endpoint):
    """Summarize video endpoint with streaming output."""

    async def initialize(self):
        """Initialize the endpoint."""
        self.asr_handle = await AanaDeploymentHandle.create("asr_deployment")
        self.llm_handle = await AanaDeploymentHandle.create("llm_deployment")
        await super().initialize()

    async def run(
        self, video: VideoInput
    ) -> AsyncGenerator[SummarizeVideoStreamEndpointOutput, None]:
        """Summarize video."""
        video_obj = await run_remote(download_video)(video_input=video)
        audio = extract_audio(video=video_obj)
        transcription = await self.asr_handle.transcribe(audio=audio)
        transcription_text = transcription["transcription"].text
        dialog = ChatDialog(
            messages=[
                ChatMessage(
                    role="system",
                    content="You are a helpful assistant that can summarize audio transcripts.",
                ),
                ChatMessage(
                    role="user",
                    content=f"Summarize the following video transcript into a list of bullet points: {transcription_text}",
                ),
            ]
        )
        async for chunk in self.llm_handle.chat_stream(dialog=dialog):
            chunk_text = chunk["text"]
            yield {"text": chunk_text}


endpoints = [
    {
        "name": "transcribe_video",
        "path": "/video/transcribe",
        "summary": "Transcribe a video",
        "endpoint_cls": TranscribeVideoEndpoint,
    },
    {
        "name": "summarize_video",
        "path": "/video/summarize",
        "summary": "Summarize a video",
        "endpoint_cls": SummarizeVideoEndpoint,
    },
    {
        "name": "summarize_video_stream",
        "path": "/video/summarize_stream",
        "summary": "Summarize a video with streaming output",
        "endpoint_cls": SummarizeVideoStreamEndpoint,
    },
]

aana_app = AanaSDK(name="summarize_video_app")

for deployment in deployments:
    aana_app.register_deployment(**deployment)

for endpoint in endpoints:
    aana_app.register_endpoint(**endpoint)

if __name__ == "__main__":
    aana_app.connect()  # Connects to the Ray cluster or starts a new one.
    aana_app.migrate()  # Runs the migrations to create the database tables.
    aana_app.deploy()  # Deploys the application.



from pathlib import Path

from pydantic_settings import BaseSettings

from aana.configs.db import DbSettings


class TestSettings(BaseSettings):
    """A pydantic model for test settings."""

    test_mode: bool = False
    use_deployment_cache: bool = False  # use cached deployment results for testing
    save_deployment_cache: bool = False  # save deployment results to cache for testing


class TaskQueueSettings(BaseSettings):
    """A pydantic model for task queue settings.

    Attributes:
        enabled (bool): Flag indicating if the task queue is enabled.
        num_workers (int): The number of workers in the task queue.
        execution_timeout (int): The maximum execution time for a task in seconds.
            After this time, if the task is still running,
            it will be considered as stuck and will be reassign to another worker.
    """

    enabled: bool = True
    num_workers: int = 4
    execution_timeout: int = 600


class Settings(BaseSettings):
    """A pydantic model for SDK settings."""

    tmp_data_dir: Path = Path("/tmp/aana_data")  # noqa: S108
    image_dir: Path = tmp_data_dir / "images"
    video_dir: Path = tmp_data_dir / "videos"
    audio_dir: Path = tmp_data_dir / "audios"
    model_dir: Path = tmp_data_dir / "models"
    num_workers: int = 2

    task_queue: TaskQueueSettings = TaskQueueSettings()

    db_config: DbSettings = DbSettings()

    test: TestSettings = TestSettings()


settings = Settings()



# ruff: noqa: S101
from importlib import resources

import pytest
from ray import serve

from aana.core.models.image import Image
from aana.tests.utils import (
    get_deployments_by_type,
    is_gpu_available,
    is_using_deployment_cache,
)


def get_expected_output(name):
    """Gets expected output for a given deployment name."""
    if name == "hf_blip2_opt_2_7b_pipeline_deployment":
        return [[{"generated_text": "the starry night by van gogh\n"}]]
    else:
        raise ValueError(f"Unknown deployment name: {name}")  # noqa: TRY003


@pytest.fixture(
    scope="function", params=get_deployments_by_type("HfPipelineDeployment")
)
def setup_hf_pipeline_deployment(app_setup, request):
    """Setup HF Pipeline deployment."""
    name, deployment = request.param
    deployments = [
        {
            "name": "hf_pipeline_deployment",
            "instance": deployment,
        }
    ]
    endpoints = []

    return name, deployment, app_setup(deployments, endpoints)


@pytest.mark.skipif(
    not is_gpu_available() and not is_using_deployment_cache(),
    reason="GPU is not available",
)
@pytest.mark.asyncio
@pytest.mark.parametrize(
    "image_name",
    ["Starry_Night.jpeg"],
)
async def test_hf_pipeline_deployments(setup_hf_pipeline_deployment, image_name):
    """Test HuggingFace Pipeline deployments."""
    name, deployment, app = setup_hf_pipeline_deployment

    handle = serve.get_app_handle("hf_pipeline_deployment")

    expected_output = get_expected_output(name)

    path = resources.path("aana.tests.files.images", image_name)
    output = await handle.call.remote(images=[str(path)])
    assert output == expected_output

    image = Image(path=path, save_on_disk=False, media_id=image_name)

    output = await handle.call.remote(images=[image])
    assert output == expected_output

    output = await handle.call.remote([image])
    assert output == expected_output

    output = await handle.call.remote(images=image)
    assert output == expected_output[0]

    output = await handle.call.remote(image)
    assert output == expected_output[0]



# ruff: noqa: S101
# Test whisper endpoints.

import time

import pytest

from aana.tests.utils import is_gpu_available, is_using_deployment_cache

TARGET = "stablediffusion2"

IMAGE_GENERATE_ENDPOINT = "/generate_image_rate_limited"
NON_RATE_LIMITED_ENDPOINT = "/generate_image"


@pytest.mark.skipif(
    not is_gpu_available() and not is_using_deployment_cache(),
    reason="GPU is not available",
)
@pytest.mark.parametrize("call_endpoint", [TARGET], indirect=True)
@pytest.mark.parametrize(
    "prompt, rate_limited_endpoint, non_rate_limited_endpoint",
    [
        (
            "Les Demoiselles d'Avignon but by Hans Holbein the Younger",
            IMAGE_GENERATE_ENDPOINT,
            NON_RATE_LIMITED_ENDPOINT,
        ),
    ],
)
def test_image_generate(
    one_request_worker,
    call_endpoint,
    prompt,
    rate_limited_endpoint,
    non_rate_limited_endpoint,
):
    """Test image generate2 endpoint. Also tests rate limiting."""
    # Generate image. Ignore output for this because it's meaningless.
    call_endpoint(
        rate_limited_endpoint, {"prompt": prompt}, ignore_expected_output=True
    )

    # Call non-rate-limited endpoint, it should not trigger TooManyRequestsException
    call_endpoint(
        non_rate_limited_endpoint, {"prompt": prompt}, ignore_expected_output=True
    )

    # Call again, it should trigger TooManyRequestsException
    call_endpoint(
        rate_limited_endpoint,
        {"prompt": prompt},
        expected_error="TooManyRequestsException",
    )

    # Wait 30s and try again
    time.sleep(30)

    # Call again, it should work again (still ignoring output).
    call_endpoint(
        rate_limited_endpoint, {"prompt": prompt}, ignore_expected_output=True
    )



# ruff: noqa: S101
import hashlib
import json
from importlib import resources

import rapidfuzz
import requests
from deepdiff.operator import BaseOperator

from aana.configs.settings import Settings
from aana.storage.op import drop_all_tables, run_alembic_migrations
from aana.tests.const import ALLOWED_LEVENSTEIN_ERROR_RATE
from aana.utils.json import jsonify


def is_gpu_available() -> bool:
    """Check if a GPU is available.

    Returns:
        bool: True if a GPU is available, False otherwise.
    """
    import torch

    # TODO: find the way to check if GPU is available without importing torch
    return torch.cuda.is_available()


def is_using_deployment_cache() -> bool:
    """Check if the deployment cache is being used.

    Returns:
        bool: True if the deployment cache is being used, False otherwise.
    """
    from aana.configs.settings import settings

    return settings.test.use_deployment_cache


def compare_texts(expected_text: str, text: str):
    """Compare two texts using Levenshtein distance.

    The error rate is allowed to be less than ALLOWED_LEVENSTEIN_ERROR_RATE.

    Args:
        expected_text (str): the expected text
        text (str): the actual text

    Raises:
        AssertionError: if the error rate is too high
    """
    dist = rapidfuzz.distance.Levenshtein.distance(text, expected_text)
    assert dist < len(expected_text) * ALLOWED_LEVENSTEIN_ERROR_RATE, (
        expected_text,
        text,
        dist,
    )


class LevenshteinOperator(BaseOperator):
    """Deepdiff operator class for Levenshtein distances."""

    def give_up_diffing(self, level, diff_instance) -> bool:
        """Short-circuit if we're certain to exceed error rate based on length."""
        dist = rapidfuzz.distance.Levenshtein.distance(level.t1, level.t2)
        if dist < len(level.t1) * ALLOWED_LEVENSTEIN_ERROR_RATE:
            return True
        return False


def call_streaming_endpoint(
    port: int, route_prefix: str, endpoint: str, data: dict
) -> list:
    """Call a streaming endpoint.

    Args:
        port (int): Port of the server.
        route_prefix (str): Route prefix of the server.
        endpoint (str): Endpoint to call.
        data (dict): Data to send to the endpoint.

    Returns:
        list: List of output chunks. If an error occurs, the list will contain
            only one element, which is the error response.
    """
    output = []
    r = requests.post(
        f"http://localhost:{port}{route_prefix}{endpoint}",
        data={"body": json.dumps(data)},
        timeout=30,
        stream=True,
    )
    for chunk in r.iter_content(chunk_size=None):
        chunk_output = json.loads(chunk.decode("utf-8"))
        output.append(chunk_output)
        if "error" in chunk_output:
            return [chunk_output]
    return output


def call_endpoint(
    target: str,
    port: int,
    route_prefix: str,
    endpoint_path: str,
    is_streaming: bool,
    data: dict,
) -> dict | list:
    """Call an endpoint.

    Args:
        target (str): the name of the target.
        port (int): Port of the server.
        route_prefix (str): Route prefix of the server.
        endpoint_path (str): Endpoint to call.
        is_streaming (bool): If True, the endpoint is a streaming endpoint.
        data (dict): Data to send to the endpoint.

    Returns:
        dict | list: Output of the endpoint. If the endpoint is a streaming endpoint, the output will be a list of output chunks.
            If the endpoint is not a streaming endpoint, the output will be a dict.
            If an error occurs, the output will be a dict with the error message.
    """
    if is_streaming:
        return call_streaming_endpoint(port, route_prefix, endpoint_path, data)
    else:
        r = requests.post(
            f"http://localhost:{port}{route_prefix}{endpoint_path}",
            data={"body": json.dumps(data)},
            timeout=30,
        )
        return r.json()


def compare_streaming_output(expected_output: list[dict], output: list[dict]):
    """Compare streaming output to expected output.

    Args:
        expected_output (list[dict]): Expected output.
        output (list[dict]): Actual output.

    Raises:
        AssertionError: if the output is different from the expected output.
    """
    # check that the output and expected output have the same length
    assert len(expected_output) == len(output), (
        len(expected_output),
        len(output),
        output,
    )

    # if error is expected or occurs, compare the errors
    if "error" in expected_output[0] or "error" in output[0]:
        # check that the expected output contains an error
        assert "error" in expected_output[0]
        # check that the output contains an error
        assert "error" in output[0]
        # check that the output and expected output have the same error
        assert expected_output[0]["error"] == output[0]["error"]
        return

    # Streaming output can come from multiple generators
    # (e.g. from transcription and captioning for video indexing)
    # so we need to make sure that we compare the output from each generator
    # to the corresponding expected output.
    # Simplest way to do this is to sort the output and expected output by the keys.
    expected_output.sort(key=lambda x: sorted(x.keys()))
    output.sort(key=lambda x: sorted(x.keys()))

    # compare the output to the expected output
    for expected, actual in zip(expected_output, output, strict=True):
        # check that the output and expected output are dicts
        assert isinstance(expected, dict)
        assert isinstance(actual, dict)
        # compare the keys
        assert expected.keys() == actual.keys()
        # compare the values (might need to update to a more intelligent comparison
        # of each value in the dict later, for now just compare the json strings with levenshtein distance)
        expected_json = json.dumps(expected, sort_keys=True)
        actual_json = json.dumps(actual, sort_keys=True)
        compare_texts(expected_json, actual_json)


def compare_output(expected_output: dict, output: dict):
    """Compare output to expected output.

    Args:
        expected_output (dict): Expected output.
        output (dict): Actual output.

    Raises:
        AssertionError: if the output is different from the expected output.
    """
    # check that the output and expected output are dicts
    assert isinstance(expected_output, dict)
    assert isinstance(output, dict)
    # if error is expected or occurs, compare the errors
    if "error" in expected_output or "error" in output:
        # check that the expected output contains an error
        assert "error" in expected_output
        # check that the output contains an error
        assert "error" in output
        # check that the output and expected output have the same error
        assert expected_output["error"] == output["error"]
        return

    # compare the keys
    assert expected_output.keys() == output.keys()
    # compare the values (might need to update to a more intelligent comparison
    # of each value in the dict later, for now just compare the json strings with levenshtein distance)
    expected_json = json.dumps(expected_output, sort_keys=True)
    actual_json = json.dumps(output, sort_keys=True)
    compare_texts(expected_json, actual_json)


def clear_database(aana_settings: Settings):
    """Clear the database.

    It drops all tables and runs alembic migrations to create the tables again.

    Args:
        aana_settings (Settings): AANA settings.
    """
    drop_all_tables(aana_settings)
    run_alembic_migrations(aana_settings)


def check_output(
    target: str,
    endpoint_path: str,
    key: str,
    output: dict | list,
    is_streaming: bool,
    ignore_expected_output: bool = False,
    expected_error: str | None = None,
):
    """Compare output with expected output.

    Args:
        target (str): the name of the target.
        endpoint_path (str): Endpoint path.
        key (str): Key of the expected output.
        output (dict | list): Output of the endpoint.
        is_streaming (bool): If True, the endpoint is a streaming endpoint.
        ignore_expected_output (bool, optional): If True, do not compare the output with the expected output. Defaults to False.
        expected_error (str | None, optional): Expected error. If not None, the output will be compared with the expected error
            and the expected output will be ignored. Defaults to None.

    Raises:
        AssertionError: if the output is different from the expected output.
    """
    # if we expect an error, then we only check the error
    if expected_error:
        if is_streaming:
            assert output[0]["error"] == expected_error, output
        else:
            assert output["error"] == expected_error, output
    # if we don't expect an error and we don't ignore the expected output, then we compare
    elif not ignore_expected_output:
        endpoint_key = endpoint_path.replace("/", "_")[
            1:
        ]  # e.g. /video/indexing -> video_indexing
        expected_output_path = resources.path(
            f"aana.tests.files.expected.endpoints.{target}",
            f"{endpoint_key}_{key}.json",
        )
        # Below block stores expected endpoint results as json files (when path does not exist yet).
        # if not expected_output_path.exists():
        #    with expected_output_path.open("w") as f:
        #        json.dump(output, f, indent=4, sort_keys=True)

        if not expected_output_path.exists():
            raise FileNotFoundError(expected_output_path)

        expected_output = json.loads(expected_output_path.read_text())
        try:
            if is_streaming:
                compare_streaming_output(expected_output, output)
            else:
                compare_output(expected_output, output)
        except AssertionError as e:
            raise AssertionError(  # noqa: TRY003
                f"Output of {endpoint_path} with key {key} is different from the expected output: {e}"
            ) from e

    # if we don't expect an error and we ignore the expected output,
    # then only check that the output does not contain an error
    else:
        if is_streaming:
            for chunk in output:
                assert "error" not in chunk, chunk
        else:
            assert "error" not in output, output


def call_and_check_endpoint(
    target: str,
    port: int,
    route_prefix: str,
    endpoint_path: str,
    data: dict,
    is_streaming: bool,
    ignore_expected_output: bool = False,
    expected_error: str | None = None,
) -> dict | list:
    """Call endpoint and compare the output with the expected output.

    Args:
        target (str): the name of the target.
        port (int): Port of the server.
        route_prefix (str): Route prefix of the server.
        endpoint_path (str): Endpoint to call.
        data (dict): Data to send to the endpoint.
        is_streaming (bool): If True, the endpoint is a streaming endpoint.
        ignore_expected_output (bool, optional): If True, do not compare the output with the expected output. Defaults to False.
        expected_error (str | None, optional): Expected error. If not None, the output will be compared with the expected error
            and the expected output will be ignored. Defaults to None.

    Returns:
        dict | list: Output of the endpoint. If the endpoint is a streaming endpoint, the output will be a list of output chunks.
            If the endpoint is not a streaming endpoint, the output will be a dict.
            If an error occurs, the output will be a dict with the error message.
    """
    data_json = jsonify(data)
    # "aana.tests.files.videos" will be resolved to a different path on different systems
    # so we need to replace it with a path that is the same on all systems
    # to make sure that the hash of the data is the same on all systems
    data_json = data_json.replace(
        str(resources.path("aana.tests.files", "")),
        "/aana/tests/files/",
    )
    data_hash = hashlib.md5(
        data_json.encode("utf-8"),
        usedforsecurity=False,
    ).hexdigest()
    output = call_endpoint(
        target=target,
        port=port,
        route_prefix=route_prefix,
        endpoint_path=endpoint_path,
        is_streaming=is_streaming,
        data=data,
    )
    check_output(
        target=target,
        endpoint_path=endpoint_path,
        key=data_hash,
        output=output,
        is_streaming=is_streaming,
        ignore_expected_output=ignore_expected_output,
        expected_error=expected_error,
    )
    return output


def get_deployments_by_type(deployment_type: str) -> list:
    """Get deployments by type from the list of all available deployments.

    Args:
        deployment_type (str): Deployment type. For example, "VLLMDeployment".

    Returns:
        list: List of deployments with the given type.
    """
    from aana.configs.deployments import available_deployments

    return [
        (name, deployment)
        for name, deployment in available_deployments.items()
        if deployment.name == deployment_type
    ]



import inspect
import pickle
from functools import wraps
from importlib import resources
from pathlib import Path
from typing import Any

import rapidfuzz

from aana.configs.settings import settings
from aana.utils.core import get_object_hash
from aana.utils.json import jsonify


def test_cache(func):  # noqa: C901
    """Decorator for caching and loading the results of a deployment function in testing mode.

    Keep in mind that this decorator only works for async functions and async generator functions.

    Use this decorator to annotate deployment functions that you want to cache in testing mode.

    There are 3 environment variables that control the behavior of the decorator:
    - TEST_MODE: set to "true" to enable testing mode
                (default is "false", should only be set to "true" if you are running tests)
    - USE_DEPLOYMENT_CACHE: set to "true" to enable cache usage
    - SAVE_DEPLOYMENT_CACHE: set to "true" to enable cache saving

    The decorator behaves differently in testing and production modes.

    In production mode, the decorator is a no-op.
    In testing mode, the behavior of the decorator is controlled by the environment variables USE_DEPLOYMENT_CACHE and SAVE_DEPLOYMENT_CACHE.

    If USE_DEPLOYMENT_CACHE is set to "true", the decorator will load the result from the cache if it exists. SAVE_DEPLOYMENT_CACHE is ignored.
    The decorator takes a hash of the deployment configuration and the function arguments and keyword arguments (args and kwargs) to locate the cache file.
    If the cache file exists, the decorator will load the result from the cache and return it.
    If the cache file does not exist, the decorator will try to find the cache file with the closest args and load the result from that cache file
    (function name and deployment configuration should match exactly, fuzzy matching only applies to args and kwargs).

    If USE_DEPLOYMENT_CACHE is set to "false", the decorator will execute the function and save the result to the cache if SAVE_DEPLOYMENT_CACHE is set to "true".
    """
    if not settings.test.test_mode:
        # If we are in production, the decorator is a no-op
        return func

    def get_cache_path(args, kwargs):
        """Get the path to the cache file."""
        self = args[0]

        func_name = func.__name__
        deployment_name = self.__class__.__name__

        config = args[0].config
        config_hash = get_object_hash(config)

        args_hash = get_object_hash({"args": args[1:], "kwargs": kwargs})

        return (
            resources.path("aana.tests.files.cache", "")
            / Path(deployment_name)
            / Path(f"{func_name}_{config_hash}_{args_hash}.pkl")
        )

    def save_cache(cache_path, cache, args, kwargs):
        """Save the cache to a file."""
        cache_obj = {
            "args": jsonify({"args": args[1:], "kwargs": kwargs}),
        }
        if "exception" in cache:
            cache_obj["exception"] = cache[
                "exception"
            ]  # if the cache contains an exception, save it
        else:
            cache_obj["cache"] = cache  # otherwise, cache the result
        cache_path.parent.mkdir(parents=True, exist_ok=True)
        cache_path.open("wb").write(pickle.dumps(cache_obj))

    def find_matching_cache(cache_path, args, kwargs):
        """Find the cache file with the closest args."""

        def get_args(path):
            cache = pickle.loads(path.open("rb").read())  # noqa: S301
            return cache["args"]

        args_str = jsonify({"args": args[1:], "kwargs": kwargs})
        pattern = cache_path.name.replace(cache_path.name.split("_")[-1], "*")
        candidate_cache_files = list(cache_path.parent.glob(pattern))

        if len(candidate_cache_files) == 0:
            raise FileNotFoundError(f"{cache_path.parent}/{pattern}")

        # find the cache with the closest args
        path = min(
            candidate_cache_files,
            key=lambda path: rapidfuzz.distance.Levenshtein.distance(
                args_str, get_args(path)
            ),
        )
        return Path(path)

    @wraps(func)
    async def wrapper(*args, **kwargs):
        """Wrapper for the deployment function."""
        cache_path = get_cache_path(args, kwargs)

        if settings.test.use_deployment_cache:
            # load from cache
            if not cache_path.exists():
                # raise FileNotFoundError(cache_path)
                cache_path = find_matching_cache(cache_path, args, kwargs)
            cache = pickle.loads(cache_path.open("rb").read())  # noqa: S301
            # raise exception if the cache contains an exception
            if "exception" in cache:
                raise cache["exception"]
            return cache["cache"]
        else:
            # execute the function
            try:
                result = await func(*args, **kwargs)
            except Exception as e:
                result = {"exception": e}
                raise
            finally:
                if settings.test.save_deployment_cache and not cache_path.exists():
                    # save to cache
                    save_cache(cache_path, result, args, kwargs)
            return result

    @wraps(func)
    async def wrapper_generator(*args, **kwargs):
        """Wrapper for the deployment generator function."""
        cache_path = get_cache_path(args, kwargs)

        if settings.test.use_deployment_cache:
            # load from cache
            if not cache_path.exists():
                # raise FileNotFoundError(cache_path)
                cache_path = find_matching_cache(cache_path, args, kwargs)

            cache = pickle.loads(cache_path.open("rb").read())  # noqa: S301
            # raise exception if the cache contains an exception
            if "exception" in cache:
                raise cache["exception"]
            for item in cache["cache"]:
                yield item
        else:
            cache = []
            try:
                # execute the function
                async for item in func(*args, **kwargs):
                    yield item
                    if settings.test.save_deployment_cache:
                        cache.append(item)
            except Exception as e:
                cache = {"exception": e}
                raise
            finally:
                if settings.test.save_deployment_cache and not cache_path.exists():
                    # save to cache
                    save_cache(cache_path, cache, args, kwargs)

    wrapper_generator.test_cache_enabled = True
    wrapper.test_cache_enabled = True

    if inspect.isasyncgenfunction(func):
        return wrapper_generator
    else:
        return wrapper


class BaseDeployment:
    """Base class for all deployments.

    We can use this class to define common methods for all deployments.
    For example, we can connect to the database here or download artifacts.
    """

    def __init__(self):
        """Inits to unconfigured state."""
        self.config = None
        self.configured = False

    async def reconfigure(self, config: dict[str, Any]):
        """Reconfigure the deployment.

        The method is called when the deployment is updated.
        """
        self.config = config
        if (
            settings.test.test_mode
            and settings.test.use_deployment_cache
            and self.check_test_cache_enabled()
        ):
            # If we are in testing mode and we want to use the cache,
            # we don't need to load the model
            self.configured = True
            return
        else:
            await self.apply_config(config)
            self.configured = True

    async def apply_config(self, config: dict[str, Any]):
        """Apply the configuration.

        Args:
            config (dict): the configuration
        """
        raise NotImplementedError

    async def get_methods(self) -> dict:
        """Returns the methods of the deployment.

        Returns:
            dict: the methods of the deployment with annotations and docstrings
        """
        cls = self.__class__
        methods = inspect.getmembers(cls, predicate=inspect.isfunction)
        methods_info = {}
        for name, method in methods:
            # Skip private methods
            if name.startswith("_"):
                continue
            # Skip non-asynchronous methods
            if not (
                inspect.iscoroutinefunction(method)
                or inspect.isasyncgenfunction(method)
            ):
                continue

            methods_info[name] = {}
            if method.__annotations__:
                methods_info[name]["annotations"] = method.__annotations__
            if method.__doc__:
                methods_info[name]["doc"] = method.__doc__
        return methods_info

    def check_test_cache_enabled(self):
        """Check if the deployment has any methods decorated with test_cache."""
        for method in self.__class__.__dict__.values():
            if callable(method) and getattr(method, "test_cache_enabled", False):
                return True
        return False



from typing import Annotated, TypedDict

import numpy as np
from pydantic import Field

from aana.api.api_generation import Endpoint
from aana.core.models.chat import Prompt
from aana.deployments.aana_deployment_handle import AanaDeploymentHandle


class ImageGenerationEndpointOutput(TypedDict):
    """Image generation endpoint output."""

    image: Annotated[
        list, Field(description="The generated image as a array of pixels.")
    ]


class ImageGenerationEndpoint(Endpoint):
    """Image generation endpoint."""

    async def initialize(self):
        """Initialize the endpoint."""
        self.image_generation_handle = await AanaDeploymentHandle.create(
            "image_generation_deployment"
        )
        await super().initialize()

    async def run(self, prompt: Prompt) -> ImageGenerationEndpointOutput:
        """Run the image generation endpoint."""
        image_generation_output = await self.image_generation_handle.generate(
            prompt=prompt,
        )
        image = np.array(image_generation_output["image"]).tolist()
        return ImageGenerationEndpointOutput(image=image)



import json
from collections.abc import AsyncGenerator
from typing import TYPE_CHECKING, Annotated, TypedDict

from pydantic import Field

from aana.api.api_generation import Endpoint
from aana.core.models.asr import (
    AsrSegments,
    AsrTranscription,
    AsrTranscriptionInfo,
)
from aana.core.models.chat import Question
from aana.core.models.media import MediaId
from aana.core.models.sampling import SamplingParams
from aana.core.models.vad import VadParams
from aana.core.models.video import VideoInput, VideoMetadata, VideoParams, VideoStatus
from aana.core.models.whisper import BatchedWhisperParams, WhisperParams
from aana.deployments.aana_deployment_handle import AanaDeploymentHandle
from aana.exceptions.db import MediaIdAlreadyExistsException, UnfinishedVideoException
from aana.exceptions.io import VideoTooLongException
from aana.integrations.external.decord import generate_frames, get_video_duration
from aana.integrations.external.yt_dlp import download_video, get_video_metadata
from aana.processors.remote import run_remote
from aana.processors.video import extract_audio, generate_combined_timeline
from aana.projects.chat_with_video.const import (
    asr_model_name,
    captioning_model_name,
    max_video_len,
)
from aana.projects.chat_with_video.utils import (
    generate_dialog,
)
from aana.storage.models.video import Status
from aana.storage.services.video import (
    check_media_id_exist,
    delete_media,
    get_video_status,
    load_video_captions,
    load_video_metadata,
    load_video_transcription,
    save_video,
    save_video_captions,
    save_video_transcription,
    update_video_status,
)

if TYPE_CHECKING:
    from aana.core.models.audio import Audio
    from aana.core.models.video import Video


class IndexVideoOutput(TypedDict):
    """The output of the transcribe video endpoint."""

    media_id: MediaId
    metadata: VideoMetadata
    transcription: AsrTranscription
    transcription_info: AsrTranscriptionInfo
    segments: AsrSegments

    captions: Annotated[list[str], Field(..., description="Captions")]
    timestamps: Annotated[
        list[float], Field(..., description="Timestamps for each caption in seconds")
    ]

    transcription_id: Annotated[int, Field(..., description="Transcription Id")]
    caption_ids: Annotated[list[int], Field(..., description="Caption Ids")]


class VideoChatEndpointOutput(TypedDict):
    """Video chat endpoint output."""

    completion: Annotated[str, Field(description="Generated text.")]


class LoadVideoMetadataOutput(TypedDict):
    """The output of the load video metadata endpoint."""

    metadata: VideoMetadata


class VideoStatusOutput(TypedDict):
    """The output of the video status endpoint."""

    status: VideoStatus


class DeleteMediaOutput(TypedDict):
    """The output of the delete media endpoint."""

    media_id: MediaId


class SimpleTranscribeVideoOutput(TypedDict):
    """The output of the simple transcribe video endpoint."""

    transcription: AsrTranscription
    transcription_info: AsrTranscriptionInfo
    segments: AsrSegments


class SimpleTranscribeVideoEndpoint(Endpoint):
    """Simple video transcription (no saving)."""

    async def initialize(self):
        """Initialize the endpoint."""
        self.asr_handle = await AanaDeploymentHandle.create("asr_deployment")

    async def run(
        self, video: VideoInput, whisper_params: WhisperParams
    ) -> AsyncGenerator[SimpleTranscribeVideoOutput, None]:
        """Transcribe video."""
        video_obj: Video = await run_remote(download_video)(video_input=video)
        audio: Audio = extract_audio(video=video_obj)

        transcription_list = []
        segments_list = []
        transcription_info_list = []
        async for whisper_output in self.asr_handle.transcribe_stream(
            audio=audio, params=whisper_params
        ):
            transcription_list.append(whisper_output["transcription"])
            segments_list.append(whisper_output["segments"])
            transcription_info_list.append(whisper_output["transcription_info"])
            yield {
                "transcription": whisper_output["transcription"],
                "segments": whisper_output["segments"],
                "info": whisper_output["transcription_info"],
            }


class IndexVideoEndpoint(Endpoint):
    """Transcribe video in chunks endpoint."""

    async def initialize(self):
        """Initialize the endpoint."""
        self.asr_handle = await AanaDeploymentHandle.create("asr_deployment")
        self.vad_handle = await AanaDeploymentHandle.create("vad_deployment")
        self.captioning_handle = await AanaDeploymentHandle.create(
            "captioning_deployment"
        )
        await super().initialize()

    async def run(
        self,
        video: VideoInput,
        video_params: VideoParams,
        whisper_params: BatchedWhisperParams,
        vad_params: VadParams,
    ) -> AsyncGenerator[IndexVideoOutput, None]:
        """Transcribe video in chunks."""
        media_id = video.media_id
        if check_media_id_exist(media_id):
            raise MediaIdAlreadyExistsException(table_name="media", media_id=video)

        video_duration = None
        if video.url is not None:
            video_metadata = get_video_metadata(video.url)
            video_duration = video_metadata.duration

        # precheck for max video length before actually download the video if possible
        if video_duration and video_duration > max_video_len:
            raise VideoTooLongException(
                video=video,
                video_len=video_duration,
                max_len=max_video_len,
            )

        video_obj: Video = await run_remote(download_video)(video_input=video)
        if video_duration is None:
            video_duration = await run_remote(get_video_duration)(video=video_obj)

        if video_duration > max_video_len:
            raise VideoTooLongException(
                video=video_obj,
                video_len=video_duration,
                max_len=max_video_len,
            )

        save_video(video=video_obj, duration=video_duration)
        yield {
            "media_id": media_id,
            "metadata": VideoMetadata(
                title=video_obj.title,
                description=video_obj.description,
                duration=video_duration,
            ),
        }

        try:
            update_video_status(media_id=media_id, status=Status.RUNNING)
            audio: Audio = extract_audio(video=video_obj)

            # TODO: Update once batched whisper PR is merged
            # vad_output = await self.vad_handle.asr_preprocess_vad(
            #     audio=audio, params=vad_params
            # )
            # vad_segments = vad_output["segments"]

            transcription_list = []
            segments_list = []
            transcription_info_list = []
            async for whisper_output in self.asr_handle.transcribe_stream(
                audio=audio, params=whisper_params
            ):
                transcription_list.append(whisper_output["transcription"])
                segments_list.append(whisper_output["segments"])
                transcription_info_list.append(whisper_output["transcription_info"])
                yield {
                    "transcription": whisper_output["transcription"],
                    "segments": whisper_output["segments"],
                    "info": whisper_output["transcription_info"],
                }
            transcription = sum(transcription_list, AsrTranscription())
            segments = sum(segments_list, AsrSegments())
            transcription_info = sum(transcription_info_list, AsrTranscriptionInfo())

            captions = []
            timestamps = []
            frame_ids = []

            async for frames_dict in run_remote(generate_frames)(
                video=video_obj, params=video_params
            ):
                if len(frames_dict["frames"]) == 0:
                    break

                timestamps.extend(frames_dict["timestamps"])
                frame_ids.extend(frames_dict["frame_ids"])

                captioning_output = await self.captioning_handle.generate_batch(
                    images=frames_dict["frames"]
                )
                captions.extend(captioning_output["captions"])

                yield {
                    "captions": captioning_output["captions"],
                    "timestamps": frames_dict["timestamps"],
                }

            save_video_transcription_output = save_video_transcription(
                model_name=asr_model_name,
                media_id=video_obj.media_id,
                transcription=transcription,
                segments=segments,
                transcription_info=transcription_info,
            )

            save_video_captions_output = save_video_captions(
                model_name=captioning_model_name,
                media_id=video_obj.media_id,
                captions=captions,
                timestamps=timestamps,
                frame_ids=frame_ids,
            )

            yield {
                "transcription_id": save_video_transcription_output["transcription_id"],
                "caption_ids": save_video_captions_output["caption_ids"],
            }
        except BaseException:
            update_video_status(media_id=media_id, status=Status.FAILED)
            raise
        else:
            update_video_status(media_id=media_id, status=Status.COMPLETED)


class VideoChatEndpoint(Endpoint):
    """Video chat endpoint."""

    async def initialize(self):
        """Initialize the endpoint."""
        self.llm_handle = await AanaDeploymentHandle.create("llm_deployment")
        await super().initialize()

    async def run(
        self, media_id: MediaId, question: Question, sampling_params: SamplingParams
    ) -> AsyncGenerator[VideoChatEndpointOutput, None]:
        """Run the video chat endpoint."""
        # check to see if video already processed
        video_status = get_video_status(media_id=media_id)
        if video_status != Status.COMPLETED:
            raise UnfinishedVideoException(
                media_id=media_id,
                status=video_status,
                message=f"The video data is not available, status: {video_status}",
            )

        load_video_transcription_output = load_video_transcription(
            media_id=media_id, model_name=asr_model_name
        )

        loaded_video_captions_output = load_video_captions(
            media_id=media_id, model_name=captioning_model_name
        )

        video_metadata = load_video_metadata(media_id=media_id)

        timeline_output = generate_combined_timeline(
            transcription_segments=load_video_transcription_output["segments"],
            captions=loaded_video_captions_output["captions"],
            caption_timestamps=loaded_video_captions_output["timestamps"],
        )
        timeline_json = json.dumps(
            timeline_output["timeline"], indent=4, separators=(",", ": ")
        )

        dialog = generate_dialog(
            metadata=video_metadata,
            timeline=timeline_json,
            question=question,
        )
        async for item in self.llm_handle.chat_stream(
            dialog=dialog, sampling_params=sampling_params
        ):
            yield {"completion": item["text"]}


class LoadVideoMetadataEndpoint(Endpoint):
    """Load video metadata endpoint."""

    async def run(self, media_id: MediaId) -> LoadVideoMetadataOutput:
        """Load video metadata."""
        video_metadata: VideoMetadata = load_video_metadata(media_id=media_id)
        return {
            "metadata": video_metadata,
        }


class GetVideoStatusEndpoint(Endpoint):
    """Get video status endpoint."""

    async def run(self, media_id: MediaId) -> VideoStatusOutput:
        """Load video metadata."""
        video_status: Status = get_video_status(media_id=media_id)
        return {
            "status": video_status,
        }


class DeleteMediaEndpoint(Endpoint):
    """Delete media endpoint."""

    async def run(self, media_id: MediaId) -> DeleteMediaOutput:
        """Delete media."""
        delete_media(media_id=media_id)
        return {"media_id": media_id}



from datetime import datetime
from typing import Any

from sqlalchemy.orm import Session

from aana.core.models.task import TaskInfo
from aana.storage.engine import engine
from aana.storage.models.task import Status as TaskStatus
from aana.storage.models.task import TaskEntity
from aana.storage.repository.task import TaskRepository


def create_task(
    endpoint: str,
    data: Any,
    priority: int = 0,
) -> str:
    """Create a task.

    Args:
        endpoint: The endpoint to which the task is assigned.
        data: Data for the task.
        priority: Priority of the task (0 is the lowest).

    Returns:
        str: The task ID.
    """
    with Session(engine) as session:
        task_repo = TaskRepository(session)
        task = TaskEntity(endpoint=endpoint, data=data, priority=priority)
        task_repo.create(task)
        return str(task.id)


def get_task_info(task_id: str) -> TaskInfo:
    """Get a task info.

    Args:
        task_id (str): The task ID.

    Returns:
        TaskStatus: The task status.

    Raises:
        NotFoundException: If the task is not found.
    """
    with Session(engine) as session:
        task_repo = TaskRepository(session)
        task = task_repo.read(task_id)
        return TaskInfo(
            id=str(task.id),
            status=task.status,
            result=task.result,
        )


def get_task(task_id: str) -> TaskEntity:
    """Get a task.

    Args:
        task_id (str): The task ID.

    Returns:
        TaskEntity: The task.

    Raises:
        NotFoundException: If the task is not found.
    """
    with Session(engine) as session:
        task_repo = TaskRepository(session)
        task = task_repo.read(task_id)
        return task


def delete_task(task_id: str) -> TaskEntity:
    """Delete a task.

    Args:
        task_id (str): The task ID.

    Returns:
        TaskEntity: The deleted task.

    Raises:
        NotFoundException: If the task is not found.
    """
    with Session(engine) as session:
        task_repo = TaskRepository(session)
        task = task_repo.delete(task_id, check=True)
        return task


def update_task_status(
    task_id: str, status: TaskStatus, progress: int | None = None, result: Any = None
):
    """Update the task status.

    Args:
        task_id (str): The task ID.
        status (TaskStatus): The new status.
        progress (int | None): The progress. If None, the progress will not be updated.
        result (Any): The result.
    """
    with Session(engine) as session:
        task_repo = TaskRepository(session)
        task = task_repo.read(task_id)
        task.status = status
        if status == TaskStatus.COMPLETED or status == TaskStatus.FAILED:
            task.completed_at = datetime.now()  # noqa: DTZ005
        if progress is not None:
            task.progress = progress
        task.result = result
        session.commit()


def get_unprocessed_tasks(limit: int | None = None) -> list[TaskEntity]:
    """Fetches all unprocessed tasks.

    The task is considered unprocessed if it is in CREATED or NOT_FINISHED state or
    in RUNNING or ASSIGNED state and the update timestamp is older
    than the execution timeout (to handle stuck tasks).

    Args:
        limit (int | None): The maximum number of tasks to fetch. If None, fetch all.

    Returns:
        list[TaskEntity]: the unprocessed tasks.
    """
    with Session(engine) as session:
        task_repo = TaskRepository(session)
        return task_repo.get_unprocessed_tasks(limit)



from transformers import BitsAndBytesConfig

from aana.core.models.sampling import SamplingParams
from aana.core.models.types import Dtype
from aana.deployments.hf_blip2_deployment import HFBlip2Config, HFBlip2Deployment
from aana.deployments.hf_pipeline_deployment import (
    HfPipelineConfig,
    HfPipelineDeployment,
)
from aana.deployments.hf_text_generation_deployment import (
    HfTextGenerationConfig,
    HfTextGenerationDeployment,
)
from aana.deployments.stablediffusion2_deployment import (
    StableDiffusion2Config,
    StableDiffusion2Deployment,
)
from aana.deployments.vad_deployment import VadConfig, VadDeployment
from aana.deployments.vllm_deployment import VLLMConfig, VLLMDeployment
from aana.deployments.whisper_deployment import (
    WhisperComputeType,
    WhisperConfig,
    WhisperDeployment,
    WhisperModelSize,
)

available_deployments = {}

vllm_llama2_7b_chat_deployment = VLLMDeployment.options(
    num_replicas=1,
    max_ongoing_requests=1000,
    ray_actor_options={"num_gpus": 0.25},
    user_config=VLLMConfig(
        model="TheBloke/Llama-2-7b-Chat-AWQ",
        dtype=Dtype.AUTO,
        quantization="awq",
        gpu_memory_reserved=13000,
        enforce_eager=True,
        default_sampling_params=SamplingParams(
            temperature=0.0, top_p=1.0, top_k=-1, max_tokens=1024
        ),
        chat_template="llama2",
    ).model_dump(mode="json"),
)

available_deployments["vllm_llama2_7b_chat_deployment"] = vllm_llama2_7b_chat_deployment

meta_llama3_8b_instruct_deployment = VLLMDeployment.options(
    num_replicas=1,
    max_ongoing_requests=1000,
    ray_actor_options={"num_gpus": 0.45},
    user_config=VLLMConfig(
        model="meta-llama/Meta-Llama-3-8B-Instruct",
        dtype=Dtype.AUTO,
        gpu_memory_reserved=30000,
        enforce_eager=True,
        default_sampling_params=SamplingParams(
            temperature=0.0, top_p=1.0, top_k=-1, max_tokens=1024
        ),
    ).model_dump(mode="json"),
)

available_deployments[
    "meta_llama3_8b_instruct_deployment"
] = meta_llama3_8b_instruct_deployment

internlm2_5_7b_chat_deployment = VLLMDeployment.options(
    num_replicas=1,
    ray_actor_options={"num_gpus": 0.45},
    user_config=VLLMConfig(
        model="internlm/internlm2_5-7b-chat",
        dtype=Dtype.AUTO,
        gpu_memory_reserved=30000,
        max_model_len=50000,
        enforce_eager=True,
        default_sampling_params=SamplingParams(
            temperature=0.0, top_p=1.0, top_k=-1, max_tokens=1024
        ),
        engine_args={"trust_remote_code": True},
    ).model_dump(mode="json"),
)

available_deployments["internlm2_5_7b_chat_deployment"] = internlm2_5_7b_chat_deployment


microsoft_phi_3_mini_instruct_deployment = VLLMDeployment.options(
    num_replicas=1,
    max_ongoing_requests=1000,
    ray_actor_options={"num_gpus": 0.25},
    user_config=VLLMConfig(
        model="microsoft/Phi-3-mini-4k-instruct",
        dtype=Dtype.AUTO,
        gpu_memory_reserved=10000,
        enforce_eager=True,
        default_sampling_params=SamplingParams(
            temperature=0.0, top_p=1.0, top_k=-1, max_tokens=1024
        ),
        engine_args={
            "trust_remote_code": True,
        },
    ).model_dump(mode="json"),
)

available_deployments[
    "microsoft_phi_3_mini_instruct_deployment"
] = microsoft_phi_3_mini_instruct_deployment

hf_blip2_opt_2_7b_deployment = HFBlip2Deployment.options(
    num_replicas=1,
    max_ongoing_requests=1000,
    ray_actor_options={"num_gpus": 0.25},
    user_config=HFBlip2Config(
        model="Salesforce/blip2-opt-2.7b",
        dtype=Dtype.FLOAT16,
        batch_size=2,
        num_processing_threads=2,
    ).model_dump(mode="json"),
)
available_deployments["hf_blip2_opt_2_7b_deployment"] = hf_blip2_opt_2_7b_deployment

whisper_medium_deployment = WhisperDeployment.options(
    num_replicas=1,
    max_ongoing_requests=1000,
    ray_actor_options={"num_gpus": 0.25},
    user_config=WhisperConfig(
        model_size=WhisperModelSize.MEDIUM,
        compute_type=WhisperComputeType.FLOAT16,
    ).model_dump(mode="json"),
)
available_deployments["whisper_medium_deployment"] = whisper_medium_deployment

stablediffusion2_deployment = StableDiffusion2Deployment.options(
    num_replicas=1,
    max_ongoing_requests=1000,
    ray_actor_options={"num_gpus": 1},
    user_config=StableDiffusion2Config(
        model="stabilityai/stable-diffusion-2",
        dtype=Dtype.FLOAT16,
    ).model_dump(mode="json"),
)
available_deployments["stablediffusion2_deployment"] = stablediffusion2_deployment

vad_deployment = VadDeployment.options(
    num_replicas=1,
    max_ongoing_requests=1000,
    ray_actor_options={"num_gpus": 0.05},
    user_config=VadConfig(
        model=(
            "https://whisperx.s3.eu-west-2.amazonaws.com/model_weights/segmentation/"
            "0b5b3216d60a2d32fc086b47ea8c67589aaeb26b7e07fcbe620d6d0b83e209ea/pytorch_model.bin"
        ),
        onset=0.5,
        sample_rate=16000,
    ).model_dump(mode="json"),
)
available_deployments["vad_deployment"] = vad_deployment

hf_blip2_opt_2_7b_pipeline_deployment = HfPipelineDeployment.options(
    num_replicas=1,
    ray_actor_options={"num_gpus": 1},
    user_config=HfPipelineConfig(
        model_id="Salesforce/blip2-opt-2.7b",
        model_kwargs={
            "quantization_config": BitsAndBytesConfig(
                load_in_8bit=False, load_in_4bit=True
            ),
        },
    ).model_dump(mode="json"),
)
available_deployments[
    "hf_blip2_opt_2_7b_pipeline_deployment"
] = hf_blip2_opt_2_7b_pipeline_deployment


hf_phi3_mini_4k_instruct_text_gen_deployment = HfTextGenerationDeployment.options(
    num_replicas=1,
    ray_actor_options={"num_gpus": 0.25},
    user_config=HfTextGenerationConfig(
        model_id="microsoft/Phi-3-mini-4k-instruct",
        model_kwargs={
            "trust_remote_code": True,
            "quantization_config": BitsAndBytesConfig(
                load_in_8bit=False, load_in_4bit=True
            ),
        },
    ).model_dump(mode="json"),
)

available_deployments[
    "hf_phi3_mini_4k_instruct_text_gen_deployment"
] = hf_phi3_mini_4k_instruct_text_gen_deployment



from aana.projects.whisper.app import aana_app

asr_deployment = aana_app.get_deployment_app("asr_deployment")
vad_deployment = aana_app.get_deployment_app("vad_deployment")
whisper_app = aana_app.get_main_app()



from aana.configs.deployments import vad_deployment, whisper_medium_deployment
from aana.projects.whisper.endpoints import (
    DeleteMediaEndpoint,
    LoadTranscriptionEndpoint,
    TranscribeVideoEndpoint,
)
from aana.sdk import AanaSDK

deployments = [
    {
        "name": "asr_deployment",
        "instance": whisper_medium_deployment,
    },
    {
        "name": "vad_deployment",
        "instance": vad_deployment,
    },
]

endpoints = [
    {
        "name": "whisper_transcribe",
        "path": "/video/transcribe",
        "summary": "Transcribe a video",
        "endpoint_cls": TranscribeVideoEndpoint,
    },
    # TODO: Update once batched whisper PR is merged
    # {
    #     "name": "whisper_transcribe_in_chunks",
    #     "path": "/video/transcribe_in_chunks",
    #     "summary": "Transcribe a video using Whisper by segmenting it into chunks",
    #     "endpoint_cls": TranscribeVideoInChunksEndpoint,
    # },
    {
        "name": "load_transcription",
        "path": "/video/get_transcription",
        "summary": "Load a transcription",
        "endpoint_cls": LoadTranscriptionEndpoint,
    },
    {
        "name": "delete_media",
        "path": "/video/delete",
        "summary": "Delete a media",
        "endpoint_cls": DeleteMediaEndpoint,
    },
]

aana_app = AanaSDK(name="whisper_app")

for deployment in deployments:
    aana_app.register_deployment(
        name=deployment["name"],
        instance=deployment["instance"],
    )

for endpoint in endpoints:
    aana_app.register_endpoint(
        name=endpoint["name"],
        path=endpoint["path"],
        summary=endpoint["summary"],
        endpoint_cls=endpoint["endpoint_cls"],
    )



from os import PathLike

from pydantic_settings import BaseSettings
from sqlalchemy.engine import Engine
from typing_extensions import TypedDict

from aana.storage.op import DbType, create_database_engine


class SQLiteConfig(TypedDict):
    """Config values for SQLite."""

    path: PathLike | str


class PostgreSQLConfig(TypedDict):
    """Config values for PostgreSQL."""

    host: str
    port: str
    user: str
    password: str
    database: str


class DbSettings(BaseSettings):
    """Database configuration."""

    datastore_type: DbType | str = DbType.SQLITE
    datastore_config: SQLiteConfig | PostgreSQLConfig = SQLiteConfig(
        path="/var/lib/aana_data"
    )
    engine: Engine | None = None

    def get_engine(self):
        """Gets engine. Each instance of DbSettings will create a max.of 1 engine."""
        if not self.engine:
            self.engine = create_database_engine(self)
        return self.engine

    def __getstate__(self):
        """Used by pickle to pickle an object."""
        # We need to remove the "engine" property because SqlAlchemy engines
        # are not picklable
        state = self.__dict__.copy()
        state.pop("engine", None)
        return state

    def __setstate__(self, state):
        """Used to restore a runtime object from pickle; the opposite of __getstate__()."""
        # We don't need to do anything special here, since the engine will be recreated
        # if needed.
        self.__dict__.update(state)



import pickle
from typing import Annotated

from pydantic import BeforeValidator, PlainSerializer

CustomConfig = Annotated[
    dict,
    PlainSerializer(lambda x: pickle.dumps(x).decode("latin1"), return_type=str),
    BeforeValidator(
        lambda x: x if isinstance(x, dict) else pickle.loads(x.encode("latin1"))  # noqa: S301
    ),
]



from aana.core.models.media import MediaId
from aana.exceptions.core import BaseException
from aana.storage.models.video import Status as VideoStatus


class NotFoundException(BaseException):
    """Raised when an item searched by id is not found."""

    def __init__(self, table_name: str, id: int | MediaId):  # noqa: A002
        """Constructor.

        Args:
            table_name (str): the name of the table being queried.
            id (int | MediaId): the id of the item to be retrieved.
        """
        super().__init__(table=table_name, id=id)
        self.table_name = table_name
        self.id = id
        self.http_status_code = 404

    def __reduce__(self):
        """Used for pickling."""
        return (self.__class__, (self.table_name, self.id))


class MediaIdAlreadyExistsException(BaseException):
    """Raised when a media_id already exists."""

    def __init__(self, table_name: str, media_id: MediaId):
        """Constructor.

        Args:
            table_name (str): the name of the table being queried.
            media_id (MediaId): the id of the item to be retrieved.
        """
        super().__init__(table=table_name, id=media_id)
        self.table_name = table_name
        self.media_id = media_id

    def __reduce__(self):
        """Used for pickling."""
        return (self.__class__, (self.table_name, self.media_id))


class UnfinishedVideoException(BaseException):
    """Exception raised when try to fetch unfinished video.

    Attributes:
        media_id (int | MediaId): The id of video.
        status (VideoStatus): The current video status.
        message (str): The error message.
    """

    def __init__(self, media_id: int | MediaId, status: VideoStatus, message: str):
        """Constructor.

        Args:
            media_id (int | MediaId): The id of video.
            status (VideoStatus): The current video status.
            message (str): The error message.
        """
        super().__init__(media_id=media_id, status=status, message=message)
        self.media_id = media_id
        self.status = status
        self.message = message

    def __reduce__(self):
        """Used for pickling."""
        return (self.__class__, (self.media_id, self.status, self.message))



import contextlib
from collections.abc import AsyncGenerator
from typing import Any

from pydantic import BaseModel, Field
from ray import serve
from vllm.engine.arg_utils import AsyncEngineArgs
from vllm.engine.async_llm_engine import AsyncLLMEngine
from vllm.inputs import TokensPrompt

from aana.core.models.base import merged_options
from aana.core.models.custom_config import CustomConfig
from aana.deployments.base_text_generation_deployment import (
    BaseTextGenerationDeployment,
    LLMOutput,
)
from aana.utils.gpu import get_gpu_memory

with contextlib.suppress(ImportError):
    from vllm.model_executor.utils import (
        set_random_seed,  # Ignore if we don't have GPU and only run on CPU with test cache
    )
from vllm.sampling_params import SamplingParams as VLLMSamplingParams
from vllm.utils import random_uuid

from aana.core.models.sampling import SamplingParams
from aana.deployments.base_deployment import test_cache
from aana.exceptions.runtime import InferenceException, PromptTooLongException


class VLLMConfig(BaseModel):
    """The configuration of the vLLM deployment.

    Attributes:
        model (str): the model name
        dtype (str): the data type (optional, default: "auto")
        quantization (str): the quantization method (optional, default: None)
        gpu_memory_reserved (float): the GPU memory reserved for the model in mb
        default_sampling_params (SamplingParams): the default sampling parameters.
        max_model_len (int): the maximum generated text length in tokens (optional, default: None)
        chat_template (str): the name of the chat template, if not provided, the chat template from the model will be used
                             but some models may not have a chat template (optional, default: None)
        enforce_eager: whether to enforce eager execution (optional, default: False)
        engine_args: extra engine arguments (optional, default: {})

    """

    model: str
    dtype: str | None = Field(default="auto")
    quantization: str | None = Field(default=None)
    gpu_memory_reserved: float
    default_sampling_params: SamplingParams = SamplingParams(
        temperature=0, max_tokens=256
    )
    max_model_len: int | None = Field(default=None)
    chat_template: str | None = Field(default=None)
    enforce_eager: bool | None = Field(default=False)
    engine_args: CustomConfig = {}


@serve.deployment
class VLLMDeployment(BaseTextGenerationDeployment):
    """Deployment to serve large language models using vLLM."""

    async def apply_config(self, config: dict[str, Any]):
        """Apply the configuration.

        The method is called when the deployment is created or updated.

        It loads the model and creates the engine.

        The configuration should contain the following keys:
        - model: the model name
        - dtype: the data type (optional, default: "auto")
        - quantization: the quantization method (optional, default: None)
        - gpu_memory_reserved: the GPU memory reserved for the model in mb
        - default_sampling_params: the default sampling parameters.
        - max_model_len: the maximum generated text length in tokens (optional, default: None)
        - chat_template: the name of the chat template (optional, default: None)
        - enforce_eager: whether to enforce eager execution (optional, default: False)
        - engine_args: extra engine arguments (optional, default: {})

        Args:
            config (dict): the configuration of the deployment
        """
        config_obj = VLLMConfig(**config)
        self.model = config_obj.model
        total_gpu_memory_bytes = get_gpu_memory()
        total_gpu_memory_mb = total_gpu_memory_bytes / 1024**2
        self.gpu_memory_utilization = (
            config_obj.gpu_memory_reserved / total_gpu_memory_mb
        )
        self.default_sampling_params: SamplingParams = (
            config_obj.default_sampling_params
        )
        self.chat_template_name = config_obj.chat_template

        args = AsyncEngineArgs(
            model=config_obj.model,
            dtype=config_obj.dtype,
            quantization=config_obj.quantization,
            enforce_eager=config_obj.enforce_eager,
            gpu_memory_utilization=self.gpu_memory_utilization,
            max_model_len=config_obj.max_model_len,
            **config_obj.engine_args,
        )

        # TODO: check if the model is already loaded.
        # If it is and none of the model parameters changed, we don't need to reload the model.

        # create the engine
        self.engine = AsyncLLMEngine.from_engine_args(args)
        self.tokenizer = self.engine.engine.tokenizer.tokenizer
        self.model_config = await self.engine.get_model_config()

    @test_cache
    async def generate_stream(
        self, prompt: str, sampling_params: SamplingParams | None = None
    ) -> AsyncGenerator[LLMOutput, None]:
        """Generate completion for the given prompt and stream the results.

        Args:
            prompt (str): the prompt
            sampling_params (SamplingParams | None): the sampling parameters

        Yields:
            LLMOutput: the dictionary with the key "text" and the generated text as the value
        """
        prompt = str(prompt)

        if sampling_params is None:
            sampling_params = SamplingParams()
        sampling_params = merged_options(self.default_sampling_params, sampling_params)

        request_id = None
        # tokenize the prompt
        prompt_token_ids = self.tokenizer.encode(prompt)

        if len(prompt_token_ids) > self.model_config.max_model_len:
            raise PromptTooLongException(
                prompt_len=len(prompt_token_ids),
                max_len=self.model_config.max_model_len,
            )

        try:
            # convert SamplingParams to VLLMSamplingParams
            sampling_params_vllm = VLLMSamplingParams(
                **sampling_params.model_dump(exclude_unset=True)
            )
            # start the request
            request_id = random_uuid()
            # set the random seed for reproducibility
            set_random_seed(42)
            results_generator = self.engine.generate(
                sampling_params=sampling_params_vllm,
                request_id=request_id,
                inputs=TokensPrompt(prompt_token_ids=prompt_token_ids),
            )

            num_returned = 0
            async for request_output in results_generator:
                text_output = request_output.outputs[0].text[num_returned:]
                yield LLMOutput(text=text_output)
                num_returned += len(text_output)
        except GeneratorExit:
            # If the generator is cancelled, we need to cancel the request
            if request_id is not None:
                await self.engine.abort(request_id)
            raise
        except Exception as e:
            raise InferenceException(model_name=self.model) from e



from datetime import datetime, timedelta
from uuid import UUID

from sqlalchemy import and_, desc, or_
from sqlalchemy.orm import Session

from aana.configs.settings import settings as aana_settings
from aana.storage.models.task import Status as TaskStatus
from aana.storage.models.task import TaskEntity
from aana.storage.repository.base import BaseRepository


class TaskRepository(BaseRepository[TaskEntity]):
    """Repository for tasks."""

    def __init__(self, session: Session):
        """Constructor."""
        super().__init__(session, TaskEntity)

    def read(self, task_id: str | UUID, check: bool = True) -> TaskEntity:
        """Reads a single task by id from the database.

        Args:
            task_id (str | UUID): ID of the task to retrieve
            check (bool): whether to raise if the entity is not found (defaults to True).

        Returns:
            The corresponding task from the database if found.

        Raises:
            NotFoundException if the entity is not found and `check` is True.
        """
        if isinstance(task_id, str):
            task_id = UUID(task_id)
        return super().read(task_id, check=check)

    def delete(self, task_id: str | UUID, check: bool = False) -> TaskEntity | None:
        """Deletes a single task by id from the database.

        Args:
            task_id (str | UUID): ID of the task to delete
            check (bool): Whether to check if the task exists before deleting

        Returns:
            The deleted task from the database if found.

        Raises:
            NotFoundException: The id does not correspond to a record in the database.
        """
        if isinstance(task_id, str):
            task_id = UUID(task_id)
        return super().delete(task_id, check)

    def get_unprocessed_tasks(self, limit: int | None = None) -> list[TaskEntity]:
        """Fetches all unprocessed tasks.

        The task is considered unprocessed if it is in CREATED or NOT_FINISHED state or
        in RUNNING or ASSIGNED state and the update timestamp is older
        than the execution timeout (to handle stuck tasks).

        Args:
            limit (int | None): The maximum number of tasks to fetch. If None, fetch all.

        Returns:
            list[TaskEntity]: the unprocessed tasks.
        """
        execution_timeout = aana_settings.task_queue.execution_timeout
        cutoff_time = datetime.now() - timedelta(seconds=execution_timeout)  # noqa: DTZ005
        tasks = (
            self.session.query(TaskEntity)
            .filter(
                or_(
                    TaskEntity.status.in_(
                        [TaskStatus.CREATED, TaskStatus.NOT_FINISHED]
                    ),
                    and_(
                        TaskEntity.status.in_(
                            [TaskStatus.RUNNING, TaskStatus.ASSIGNED]
                        ),
                        TaskEntity.update_ts <= cutoff_time,
                    ),
                )
            )
            .order_by(desc(TaskEntity.priority), TaskEntity.create_ts)
            .limit(limit)
            .all()
        )
        return tasks



from __future__ import annotations  # Let classes use themselves in type annotations

import typing

from sqlalchemy import CheckConstraint, Column, Float, ForeignKey, Integer, String
from sqlalchemy.orm import relationship

from aana.storage.models.base import BaseEntity, TimeStampEntity
from aana.storage.types import MediaIdSqlType

if typing.TYPE_CHECKING:
    from aana.core.models.captions import Caption
    from aana.core.models.media import MediaId


class CaptionEntity(BaseEntity, TimeStampEntity):
    """ORM model for video captions."""

    __tablename__ = "captions"

    id = Column(Integer, autoincrement=True, primary_key=True)
    model = Column(
        String, nullable=False, comment="Name of model used to generate the caption"
    )
    media_id = Column(
        MediaIdSqlType,
        ForeignKey("video.id"),
        nullable=False,
        comment="Foreign key to video table",
    )
    frame_id = Column(
        Integer,
        CheckConstraint("frame_id >= 0", "frame_id_positive"),
        comment="The 0-based frame id of video for caption",
    )
    caption = Column(String, comment="Frame caption")
    timestamp = Column(
        Float,
        CheckConstraint("timestamp >= 0", name="timestamp_positive"),
        comment="Frame timestamp in seconds",
    )

    video = relationship("VideoEntity", back_populates="captions", uselist=False)

    @classmethod
    def from_caption_output(
        cls,
        model_name: str,
        media_id: MediaId,
        frame_id: int,
        frame_timestamp: float,
        caption: Caption,
    ) -> CaptionEntity:
        """Converts a Caption pydantic model to a CaptionEntity."""
        return CaptionEntity(
            model=model_name,
            media_id=media_id,
            frame_id=frame_id,
            caption=str(caption),
            timestamp=frame_timestamp,
        )



# ruff: noqa: S101
# Test whisper endpoints.

from importlib import resources

import pytest

from aana.tests.utils import is_gpu_available, is_using_deployment_cache

TARGET = "whisper"

VIDEO_TRANSCRIBE_ENDPOINT = "/video/transcribe"
VIDEO_GET_TRANSCRIPTION_ENDPOINT = "/video/get_transcription"
VIDEO_DELETE_ENDPOINT = "/video/delete"
# TODO: Update once batched whisper PR is merged
# VIDEO_TRANSCRIBE_BATCH_ENDPOINT = "/video/transcribe_in_chunks"


@pytest.mark.skipif(
    not is_gpu_available() and not is_using_deployment_cache(),
    reason="GPU is not available",
)
@pytest.mark.parametrize("call_endpoint", [TARGET], indirect=True)
@pytest.mark.parametrize(
    "video, whisper_params, transcription_endpoint",
    [
        (
            {
                "path": str(
                    resources.path("aana.tests.files.videos", "physicsworks.webm")
                ),
                "media_id": "physicsworks.webm",
            },
            {"temperature": 0.0},
            VIDEO_TRANSCRIBE_ENDPOINT,
        ),
        # TODO: Update once batched whisper PR is merged
        # (
        #     {
        #         "path": str(
        #             resources.path("aana.tests.files.videos", "physicsworks.webm")
        #         ),
        #         "media_id": "physicsworks.webm_batched",
        #     },
        #     {"temperature": 0.0},
        #     VIDEO_TRANSCRIBE_BATCH_ENDPOINT,
        # ),
        # (
        #     {
        #         "path": str(
        #             resources.path("aana.tests.files.audios", "physicsworks.wav")
        #         ),
        #         "media_id": "physicsworks.wav_batched",
        #     },
        #     {"temperature": 0.0},
        #     VIDEO_TRANSCRIBE_BATCH_ENDPOINT,
        # ),
        (
            {
                "path": str(
                    resources.path("aana.tests.files.audios", "physicsworks.wav")
                ),
                "media_id": "physicsworks.wav",
            },
            {"temperature": 0.0},
            VIDEO_TRANSCRIBE_ENDPOINT,
        ),
    ],
)
def test_video_transcribe(call_endpoint, video, whisper_params, transcription_endpoint):
    """Test video transcribe endpoint."""
    media_id = video["media_id"]

    # transcribe video
    call_endpoint(
        transcription_endpoint,
        {"video": video, "whisper_params": whisper_params},
    )

    # load transcription
    call_endpoint(
        VIDEO_GET_TRANSCRIPTION_ENDPOINT,
        {"media_id": media_id},
    )

    # try to transcribe video again, it should fail with MediaIdAlreadyExistsException
    call_endpoint(
        transcription_endpoint,
        {"video": video, "whisper_params": whisper_params},
        expected_error="MediaIdAlreadyExistsException",
    )

    # delete video
    call_endpoint(
        VIDEO_DELETE_ENDPOINT,
        {"media_id": media_id},
        ignore_expected_output=True,
    )

    # try to load transcription again, it should fail with NotFoundException
    call_endpoint(
        VIDEO_GET_TRANSCRIPTION_ENDPOINT,
        {"media_id": media_id},
        expected_error="NotFoundException",
    )

    # try to delete video again, it should fail with NotFoundException
    call_endpoint(
        VIDEO_DELETE_ENDPOINT,
        {"media_id": media_id},
        expected_error="NotFoundException",
    )

    # transcribe video again after deleting it
    call_endpoint(
        transcription_endpoint,
        {"video": video, "whisper_params": whisper_params},
    )

    # delete video
    call_endpoint(
        VIDEO_DELETE_ENDPOINT,
        {"media_id": media_id},
        ignore_expected_output=True,
    )

    # no found exception for deleted transcription
    call_endpoint(
        VIDEO_GET_TRANSCRIPTION_ENDPOINT,
        {"media_id": media_id},
        expected_error="NotFoundException",
    )



# ruff: noqa: S101
import json
from collections import defaultdict
from importlib import resources
from pathlib import Path

import pytest
from deepdiff import DeepDiff
from ray import serve

from aana.core.models.audio import Audio
from aana.core.models.base import pydantic_to_dict
from aana.core.models.vad import VadSegment
from aana.core.models.whisper import BatchedWhisperParams, WhisperParams
from aana.tests.utils import (
    LevenshteinOperator,
    get_deployments_by_type,
    is_gpu_available,
    is_using_deployment_cache,
)

EPSILON = 0.01


def compare_transcriptions(expected_transcription, transcription):
    """Compare two transcriptions.

    Texts and words are compared using Levenshtein distance.

    Args:
        expected_transcription (dict): the expected transcription
        transcription (dict): the actual transcription
    Raises:
        AssertionError: if transcriptions differ too much
    """
    custom_operators = [LevenshteinOperator([r"\['text'\]$", r"\['word'\]$"])]

    diff = DeepDiff(
        expected_transcription,
        transcription,
        math_epsilon=EPSILON,
        ignore_numeric_type_changes=True,
        custom_operators=custom_operators,
    )
    assert not diff, diff


@pytest.fixture(scope="function", params=get_deployments_by_type("WhisperDeployment"))
def setup_whisper_deployment(app_setup, request):
    """Setup whisper deployment."""
    name, deployment = request.param
    deployments = [
        {
            "name": "whisper_deployment",
            "instance": deployment,
        }
    ]
    endpoints = []

    return name, deployment, app_setup(deployments, endpoints)


@pytest.mark.skipif(
    not is_gpu_available() and not is_using_deployment_cache(),
    reason="GPU is not available",
)
@pytest.mark.asyncio
@pytest.mark.parametrize("audio_file", ["squirrel.wav", "physicsworks.wav"])
async def test_whisper_deployment(setup_whisper_deployment, audio_file):
    """Test whisper deployment."""
    name, deployment, app = setup_whisper_deployment

    handle = serve.get_app_handle("whisper_deployment")

    model_size = deployment.user_config["model_size"]
    audio_file_name = Path(audio_file).stem
    expected_output_path = resources.path(
        f"aana.tests.files.expected.whisper.{model_size}", f"{audio_file_name}.json"
    )
    assert (
        expected_output_path.exists()
    ), f"Expected output not found: {expected_output_path}"
    with Path(expected_output_path) as path, path.open() as f:
        expected_output = json.load(f)

    # Test transcribe method
    path = resources.path("aana.tests.files.audios", audio_file)
    assert path.exists(), f"Audio not found: {path}"
    audio = Audio(path=path, media_id=audio_file)

    output = await handle.transcribe.remote(
        audio=audio, params=WhisperParams(word_timestamps=True, temperature=0.0)
    )
    output = pydantic_to_dict(output)

    compare_transcriptions(expected_output, output)

    # Test transcribe_stream method)
    stream = handle.options(stream=True).transcribe_stream.remote(
        audio=audio, params=WhisperParams(word_timestamps=True, temperature=0.0)
    )

    # Combine individual segments and compare with the final dict
    grouped_dict = defaultdict(list)
    transcript = ""
    async for chunk in stream:
        output = pydantic_to_dict(chunk)
        transcript += output["transcription"]["text"]
        grouped_dict["segments"].extend(output.get("segments", []))

    grouped_dict["transcription"] = {"text": transcript}
    grouped_dict["transcription_info"] = output.get("transcription_info")
    compare_transcriptions(expected_output, dict(grouped_dict))

    # Test transcribe_batch method

    # Test transcribe_in_chunks method: Note that the expected asr output is different
    # TODO: Update once batched whisper PR is merged
    # expected_batched_output_path = resources.path(
    #     f"aana.tests.files.expected.whisper.{model_size}",
    #     f"{audio_file_name}_batched.json",
    # )
    # assert (
    #     expected_batched_output_path.exists()
    # ), f"Expected output not found: {expected_batched_output_path}"
    # with Path(expected_batched_output_path) as path, path.open() as f:
    #     expected_output_batched = json.load(f)

    # # Get expected vad segments
    # vad_path = resources.path(
    #     "aana.tests.files.expected.vad", f"{audio_file_name}_vad.json"
    # )
    # assert vad_path.exists(), f"vad expected predictions not found: {vad_path}"

    # with Path(vad_path) as path, path.open() as f:
    #     expected_output_vad = json.load(f)

    # final_input = [
    #     VadSegment(time_interval=seg["time_interval"], segments=seg["segments"])
    #     for seg in expected_output_vad["segments"]
    # ]

    # batched_stream = handle.options(stream=True).transcribe_in_chunks.remote(
    #     audio=audio,
    #     segments=final_input,
    #     batch_size=16,
    #     params=BatchedWhisperParams(temperature=0.0),
    # )

    # # Combine individual segments and compare with the final dict
    # transcript = ""
    # grouped_dict = defaultdict(list)
    # async for chunk in batched_stream:
    #     output = pydantic_to_dict(chunk)
    #     transcript += output["transcription"]["text"]
    #     grouped_dict["segments"].extend(output.get("segments", []))

    # grouped_dict["transcription"] = {"text": transcript}
    # grouped_dict["transcription_info"] = output.get("transcription_info")

    # compare_transcriptions(
    #     expected_output_batched,
    #     dict(grouped_dict),
    # )



# ruff: noqa: S101
from importlib import resources

import pytest
from ray import serve

from aana.core.models.image import Image
from aana.tests.utils import (
    compare_texts,
    get_deployments_by_type,
    is_gpu_available,
    is_using_deployment_cache,
)


@pytest.fixture(scope="function", params=get_deployments_by_type("HFBlip2Deployment"))
def setup_hf_blip2_deployment(app_setup, request):
    """Setup HF BLIP2 deployment."""
    name, deployment = request.param
    deployments = [
        {
            "name": "blip2_deployment",
            "instance": deployment,
        }
    ]
    endpoints = []

    return name, deployment, app_setup(deployments, endpoints)


@pytest.mark.skipif(
    not is_gpu_available() and not is_using_deployment_cache(),
    reason="GPU is not available",
)
@pytest.mark.asyncio
@pytest.mark.parametrize(
    "image_name, expected_text",
    [
        (
            "Starry_Night.jpeg",
            "the starry night by vincent van gogh, 1884-1890, oil on canvas, 48 x 48 in, gilded frame, signed and dated",
        )
    ],
)
async def test_hf_blip2_deployments(
    setup_hf_blip2_deployment, image_name, expected_text
):
    """Test HuggingFace BLIP2 deployments."""
    handle = serve.get_app_handle("blip2_deployment")

    path = resources.path("aana.tests.files.images", image_name)
    image = Image(path=path, save_on_disk=False, media_id=image_name)

    output = await handle.generate.remote(image=image)
    caption = output["caption"]
    compare_texts(expected_text, caption)

    images = [image] * 8

    output = await handle.generate_batch.remote(images=images)
    captions = output["captions"]

    assert len(captions) == 8
    for caption in captions:
        compare_texts(expected_text, caption)



from sqlalchemy.orm import Session

from aana.storage.models import VideoEntity
from aana.storage.repository.base import BaseRepository


class VideoRepository(BaseRepository[VideoEntity]):
    """Repository for media files."""

    def __init__(self, session: Session):
        """Constructor."""
        super().__init__(session, VideoEntity)



# ruff: noqa: S101
import pytest
from sqlalchemy.orm import Session

from aana.storage.models import (
    CaptionEntity,
    MediaEntity,
    TranscriptEntity,
    VideoEntity,
)
from aana.storage.repository.caption import CaptionRepository
from aana.storage.repository.media import MediaRepository
from aana.storage.repository.transcript import TranscriptRepository
from aana.storage.repository.video import VideoRepository


@pytest.fixture
def mocked_session(mocker):
    """Creates a mocked sqlalchemy.Session."""
    session = mocker.MagicMock(spec=Session)
    # Emulate the behavior of the empty database.
    session.query.return_value.filter_by.return_value.first.return_value = None
    session.query.return_value.get.return_value = None
    return session


def test_create_media(mocked_session):
    """Tests that media creation behaves as expected."""
    repo = MediaRepository(mocked_session)
    media_type = "video"
    media_id = "foo"
    media = MediaEntity(id=media_id, media_type=media_type)
    media2 = repo.create(media)

    # We need an integration test to ensure that that the id gets set
    # on creation, because the mocked version won't set it.
    assert media2 == media
    assert media2.media_type == media_type
    assert media2.id == media_id

    mocked_session.add.assert_called_once_with(media)
    mocked_session.commit.assert_called_once()


def test_create_media_with_video(mocked_session):
    """Tests that video propery is se on media and vice-versa."""
    media_repo = MediaRepository(mocked_session)
    video_repo = VideoRepository(mocked_session)
    media_type = "video"
    media_id = "foo"
    media = MediaEntity(id=media_id, media_type=media_type)
    video = VideoEntity(media=media)
    media.video = video
    media2 = media_repo.create(media)
    mocked_session.add.assert_called_with(media2)
    video2 = video_repo.create(video)
    mocked_session.add.assert_called_with(video2)

    assert media2.video == video2
    assert video2.media == media2


def test_create_caption(mocked_session):
    """Tests caption creation."""
    repo = CaptionRepository(mocked_session)
    media_id = "foo"
    media_type = "video"
    video_duration = 500.25
    model_name = "no_model"
    caption_text = "This is the right caption text."
    frame_id = 32767
    timestamp = 327.6
    _ = MediaEntity(id=media_id, media_type=media_type)
    _ = VideoEntity(id=media_id, duration=video_duration)

    caption = CaptionEntity(
        media_id=media_id,
        model=model_name,
        frame_id=frame_id,
        caption=caption_text,
        timestamp=timestamp,
    )
    caption2 = repo.create(caption)

    # See above
    assert caption2.media_id == media_id
    assert caption2.model == model_name
    assert caption2.frame_id == frame_id
    assert caption2.caption == caption_text
    assert caption2.timestamp == timestamp

    mocked_session.add.assert_called_once_with(caption)
    mocked_session.commit.assert_called_once()


def test_create_transcript(mocked_session):
    """Tests transcript creation."""
    repo = TranscriptRepository(mocked_session)
    media_id = "foo"
    duration = 500.25
    media_type = "video"
    model_name = "no_model"
    transcript_text = "This is the right transcript text."
    segments = "This is a segments string."
    language = "en"
    language_confidence = 0.5
    media = MediaEntity(id=media_id, media_type=media_type)
    _ = VideoEntity(id=media_id, duration=duration)
    transcript = TranscriptEntity(
        media_id=media_id,
        model=model_name,
        transcript=transcript_text,
        segments=segments,
        language=language,
        language_confidence=language_confidence,
    )
    transcript2 = repo.create(transcript)

    # See above
    assert transcript2.media_id == media.id
    assert transcript2.model == model_name
    assert transcript2.transcript == transcript_text
    assert transcript2.segments == segments
    assert transcript2.language == language
    assert transcript2.language_confidence == language_confidence

    mocked_session.add.assert_called_once_with(transcript)
    mocked_session.commit.assert_called_once()



from typing import Any, TypedDict

import PIL
import torch
from diffusers import EulerDiscreteScheduler, StableDiffusionPipeline
from pydantic import BaseModel, Field
from ray import serve

from aana.core.models.chat import Prompt
from aana.core.models.types import Dtype
from aana.deployments.base_deployment import BaseDeployment, test_cache


class StableDiffusion2Output(TypedDict):
    """Output class for the StableDiffusion2 deployment."""

    image: PIL.Image.Image


class StableDiffusion2Config(BaseModel):
    """The configuration for the Stable Diffusion 2 deployment.

    Attributes:
        model (str): the model ID on HuggingFace
        dtype (str): the data type (optional, default: "auto"), one of "auto", "float32", "float16"
    """

    model: str
    dtype: Dtype = Field(default=Dtype.AUTO)


@serve.deployment
class StableDiffusion2Deployment(BaseDeployment):
    """Stable Diffusion 2 deployment."""

    async def apply_config(self, config: dict[str, Any]):
        """Apply the configuration.

        The method is called when the deployment is created or updated.

        It loads the model and scheduler from HuggingFace.

        The configuration should conform to the StableDiffusion2Config schema.
        """
        config_obj = StableDiffusion2Config(**config)

        # Load the model and processor from HuggingFace
        self.model_id = config_obj.model
        self.dtype = config_obj.dtype
        if self.dtype == Dtype.INT8:
            self.torch_dtype = Dtype.FLOAT16.to_torch()
        else:
            self.torch_dtype = self.dtype.to_torch()
        self.device = "cuda" if torch.cuda.is_available() else "cpu"
        self.model = StableDiffusionPipeline.from_pretrained(
            self.model_id,
            torch_dtype=self.torch_dtype,
            scheduler=EulerDiscreteScheduler.from_pretrained(
                self.model_id, subfolder="scheduler"
            ),
            device_map="auto",
        )

        self.model.to(self.device)

    @test_cache
    async def generate(self, prompt: Prompt) -> StableDiffusion2Output:
        """Runs the model on a given prompt and returns the first output.

        Arguments:
            prompt (Prompt): the prompt to the model.

        Returns:
            StableDiffusion2Output: a dictionary with one key containing the result
        """
        image = self.model(str(prompt)).images[0]
        return {"image": image}



# ruff: noqa: S101

from importlib import resources
from tempfile import NamedTemporaryFile
from unittest.mock import patch

import pytest
from sqlalchemy.orm import Session

from aana.configs.settings import settings
from aana.core.models.video import Video
from aana.exceptions.db import NotFoundException
from aana.storage.op import run_alembic_migrations
from aana.storage.repository.media import MediaRepository
from aana.storage.repository.video import VideoRepository
from aana.storage.services.video import delete_media, save_video


def test_save_video():
    """Tests saving a video against an in-memory SQLite database."""
    # Ensure the tmp data directory exists
    settings.tmp_data_dir.mkdir(parents=True, exist_ok=True)
    # Use a temp file and change the settings to use it for the SQLite data path.
    with NamedTemporaryFile(
        dir=settings.tmp_data_dir,
    ) as tmp:
        settings.db_config.datastore_config["path"] = tmp.name
        # Reset the engine if it has been created
        del settings.db_config.engine
        settings.db_config.engine = None

        run_alembic_migrations(settings)

        # We also have to patch the "engine" variable in aana.utils.db
        # because it gets set before the test is run due to import dependencies.
        # TODO: We should stop using the engine as an importable name
        with patch(
            "aana.storage.services.video.engine", settings.db_config.get_engine()
        ):
            media_id = "foobar"
            duration = 550.25
            path = resources.path("aana.tests.files.videos", "squirrel.mp4")
            video = Video(path=path, media_id=media_id)
            result = save_video(video, duration)

            assert result["media_id"] == media_id

            # Check that saved video is now available from repo
            with Session(settings.db_config.get_engine()) as session:
                video_repo = VideoRepository(session)
                video_entity = video_repo.read(media_id)

                assert video_entity
                # Check that video has a media, that media exists
                # and that media has a video set
                media_repo = MediaRepository(session)
                media_entity = media_repo.read(media_id)
                assert media_entity.video == video_entity

            # Use a second session here because deleted objects sometimes linger
            # as long as a session persists.
            with Session(settings.db_config.get_engine()) as session:
                # Check that delete function works
                result = delete_media(media_id)
                with pytest.raises(NotFoundException):
                    _ = media_repo.read(media_id)
                with pytest.raises(NotFoundException):
                    _ = video_repo.read(media_id)



from aana.configs.deployments import (
    vllm_llama2_7b_chat_deployment,
)
from aana.projects.llama2.endpoints import (
    LlmChatEndpoint,
    LlmChatStreamEndpoint,
    LlmGenerateEndpoint,
    LlmGenerateStreamEndpoint,
)
from aana.sdk import AanaSDK

deployments = [
    {
        "name": "llm_deployment",
        "instance": vllm_llama2_7b_chat_deployment,
    },
]

endpoints = [
    {
        "name": "llm_generate",
        "path": "/llm/generate",
        "summary": "Generate text using LLaMa2 7B Chat",
        "endpoint_cls": LlmGenerateEndpoint,
    },
    {
        "name": "llm_generate_stream",
        "path": "/llm/generate_stream",
        "summary": "Generate text using LLaMa2 7B Chat (streaming)",
        "endpoint_cls": LlmGenerateStreamEndpoint,
    },
    {
        "name": "llm_chat",
        "path": "/llm/chat",
        "summary": "Chat with LLaMa2 7B Chat",
        "endpoint_cls": LlmChatEndpoint,
    },
    {
        "name": "llm_chat_stream",
        "path": "/llm/chat_stream",
        "summary": "Chat with LLaMa2 7B Chat (streaming)",
        "endpoint_cls": LlmChatStreamEndpoint,
    },
]

aana_app = AanaSDK(name="llama2_app")

for deployment in deployments:
    aana_app.register_deployment(
        name=deployment["name"],
        instance=deployment["instance"],
    )

for endpoint in endpoints:
    aana_app.register_endpoint(
        name=endpoint["name"],
        path=endpoint["path"],
        summary=endpoint["summary"],
        endpoint_cls=endpoint["endpoint_cls"],
    )



# ruff: noqa: S101
# Test chat with video endpoints.

from importlib import resources

import pytest

from aana.tests.utils import is_gpu_available, is_using_deployment_cache

TARGET = "chat_with_video"

VIDEO_INDEX_ENDPOINT = "/video/index_stream"
VIDEO_METADATA_ENDPOINT = "/video/metadata"
VIDEO_CHAT_ENDPOINT = "/video/chat_stream"
VIDEO_STATUS_ENDPOINT = "/video/status"
VIDEO_DELETE_ENDPOINT = "/video/delete"


@pytest.mark.skipif(
    not is_gpu_available() and not is_using_deployment_cache(),
    reason="GPU is not available",
)
@pytest.mark.parametrize("call_endpoint", [TARGET], indirect=True)
@pytest.mark.parametrize(
    "video, whisper_params",
    [
        (
            {
                "url": "https://mobius-public.s3.eu-west-1.amazonaws.com/squirrel.mp4",
                "media_id": "squirrel.mp4",
            },
            {"temperature": 0.0},
        ),
        (
            {
                "path": str(
                    resources.path("aana.tests.files.videos", "physicsworks.webm")
                ),
                "media_id": "physicsworks.webm",
            },
            {"temperature": 0.0},
        ),
        (
            {
                "path": str(
                    resources.path("aana.tests.files.videos", "physicsworks_audio.webm")
                ),
                "media_id": "physicsworks_audio.webm",
            },
            {"temperature": 0.0},
        ),
    ],
)
def test_chat_with_video(call_endpoint, video, whisper_params):
    """Test chat with video endpoint."""
    media_id = video["media_id"]

    call_endpoint(
        VIDEO_INDEX_ENDPOINT,
        {"video": video, "whisper_params": whisper_params},
    )

    # if we try to index the same video again, we should get an error MediaIdAlreadyExistsException
    call_endpoint(
        VIDEO_INDEX_ENDPOINT,
        {"video": video, "whisper_params": whisper_params},
        expected_error="MediaIdAlreadyExistsException",
    )

    # load video metadata
    call_endpoint(
        VIDEO_METADATA_ENDPOINT,
        {"media_id": media_id},
    )

    # get video status
    call_endpoint(
        VIDEO_STATUS_ENDPOINT,
        {"media_id": media_id},
    )

    # delete video
    call_endpoint(
        VIDEO_DELETE_ENDPOINT,
        {"media_id": media_id},
        ignore_expected_output=True,
    )

    # get video status
    call_endpoint(
        VIDEO_STATUS_ENDPOINT,
        {"media_id": media_id},
        expected_error="NotFoundException",
    )

    # after deleting the video video metadata should not be available
    call_endpoint(
        VIDEO_METADATA_ENDPOINT,
        {"media_id": media_id},
        expected_error="NotFoundException",
    )

    # after deleting the video, we should be able to index it again
    call_endpoint(
        VIDEO_INDEX_ENDPOINT,
        {"video": video, "whisper_params": whisper_params},
    )

    # load video metadata
    call_endpoint(
        VIDEO_METADATA_ENDPOINT,
        {"media_id": media_id},
    )

    # chat with video
    question = "Summarize the video"

    call_endpoint(
        VIDEO_CHAT_ENDPOINT,
        {"media_id": media_id, "question": question},
    )

    # delete video
    call_endpoint(
        VIDEO_DELETE_ENDPOINT,
        {"media_id": media_id},
        ignore_expected_output=True,
    )

    # after deleting the video, we should not be able to chat with it
    call_endpoint(
        VIDEO_CHAT_ENDPOINT,
        {"media_id": media_id, "question": question},
        expected_error="NotFoundException",
    )


@pytest.mark.parametrize(
    "endpoint, data",
    [
        (VIDEO_METADATA_ENDPOINT, {}),
        (VIDEO_CHAT_ENDPOINT, {}),
        (VIDEO_CHAT_ENDPOINT, {"media_id": "squirrel.mp4"}),
        (VIDEO_CHAT_ENDPOINT, {"question": "Summarize the video"}),
        (VIDEO_INDEX_ENDPOINT, {}),
        (VIDEO_DELETE_ENDPOINT, {}),
    ],
)
@pytest.mark.parametrize("call_endpoint", [TARGET], indirect=True)
def test_missing_params(call_endpoint, endpoint, data):
    """Test missing params."""
    call_endpoint(
        endpoint,
        data,
        expected_error="ValidationError",
    )



import asyncio
from collections.abc import AsyncGenerator
from queue import Empty
from threading import Thread
from typing import Any

import torch
import transformers
from pydantic import BaseModel
from ray import serve
from transformers import (
    AutoModelForCausalLM,
    AutoTokenizer,
    TextIteratorStreamer,
)

from aana.core.models.base import merged_options
from aana.core.models.sampling import SamplingParams
from aana.deployments.base_deployment import test_cache
from aana.deployments.base_text_generation_deployment import (
    BaseTextGenerationDeployment,
    LLMOutput,
)
from aana.deployments.hf_pipeline_deployment import CustomConfig
from aana.exceptions.runtime import InferenceException, PromptTooLongException


class HfTextGenerationConfig(BaseModel):
    """The configuration for the Hugging Face text generation deployment.

    Attributes:
        model_id (str): the model ID on Hugging Face
        model_kwargs (dict): the model keyword arguments
        default_sampling_params (SamplingParams): the default sampling parameters
        chat_template (str): the name of the chat template (optional, default: None)
    """

    model_id: str
    model_kwargs: CustomConfig = {}
    default_sampling_params: SamplingParams = SamplingParams(
        temperature=0, max_tokens=256
    )
    chat_template: str | None = None


async def async_streamer_adapter(streamer):
    """Adapt the TextIteratorStreamer to an async generator."""
    while True:
        try:
            for item in streamer:
                yield item
            break
        except Empty:
            # wait for the next item
            await asyncio.sleep(0.01)


@serve.deployment
class HfTextGenerationDeployment(BaseTextGenerationDeployment):
    """Deployment to serve Hugging Face text generation models."""

    async def apply_config(self, config: dict[str, Any]):
        """Apply the configuration.

        The method is called when the deployment is created or updated.

        It loads the model and tokenizer from HuggingFace.

        The configuration should conform to the HfTextGenerationConfig schema.
        """
        config_obj = HfTextGenerationConfig(**config)
        self.model_id = config_obj.model_id
        self.model_kwargs = config_obj.model_kwargs
        self.default_sampling_params = config_obj.default_sampling_params

        self.model_kwargs["device_map"] = self.model_kwargs.get("device_map", "auto")
        self.model = AutoModelForCausalLM.from_pretrained(
            self.model_id, **self.model_kwargs
        )
        self.tokenizer = AutoTokenizer.from_pretrained(self.model_id)
        self.chat_template_name = config_obj.chat_template

    @test_cache
    async def generate_stream(
        self, prompt: str, sampling_params: SamplingParams | None = None
    ) -> AsyncGenerator[LLMOutput, None]:
        """Generate text stream.

        Args:
            prompt (str): the prompt
            sampling_params (SamplingParams | None): the sampling parameters

        Yields:
            LLMOutput: the generated text
        """
        # Set the seed to make the results reproducible
        transformers.set_seed(42)

        prompt = str(prompt)

        if sampling_params is None:
            sampling_params = SamplingParams()
        sampling_params = merged_options(self.default_sampling_params, sampling_params)

        prompt_input = self.tokenizer(
            prompt, return_tensors="pt", add_special_tokens=False
        )

        if prompt_input["input_ids"].shape[-1] > self.tokenizer.model_max_length:
            raise PromptTooLongException(
                prompt_len=prompt_input["input_ids"].shape[-1],
                max_len=self.tokenizer.model_max_length,
            )

        try:
            with torch.no_grad():
                prompt_input.to(self.model.device)
                streamer = TextIteratorStreamer(
                    self.tokenizer,
                    timeout=0,
                    skip_prompt=True,
                    skip_special_tokens=True,
                )
                generation_kwargs = dict(
                    prompt_input,
                    streamer=streamer,
                    max_new_tokens=sampling_params.max_tokens,
                    top_k=sampling_params.top_k,
                    top_p=sampling_params.top_p,
                    temperature=sampling_params.temperature,
                    num_return_sequences=1,
                    eos_token_id=self.tokenizer.eos_token_id,
                )
                if sampling_params.temperature == 0:
                    generation_kwargs["do_sample"] = False
                    generation_kwargs["temperature"] = None
                generation_thread = Thread(
                    target=self.model.generate, kwargs=generation_kwargs
                )
                generation_thread.start()

                async_streamer = async_streamer_adapter(streamer)
                async for new_text in async_streamer:
                    yield LLMOutput(text=new_text)

                # clean up
                prompt_input.to("cpu")
                del streamer
        except Exception as e:
            raise InferenceException(model_name=self.model_id) from e



from aana.core.models.chat import ChatDialog, ChatMessage, Question
from aana.core.models.video import VideoMetadata


def generate_dialog(
    metadata: VideoMetadata,
    timeline: str,
    question: Question,
) -> ChatDialog:
    """Generates a dialog from the metadata and timeline of a video.

    Args:
        metadata (VideoMetadata): the metadata of the video
        timeline (str): the timeline of the video
        question (Question): the question to ask

    Returns:
        ChatDialog: the generated dialog
    """
    system_prompt_preamble = """You are a helpful, respectful, and honest assistant. Always answer as helpfully as possible, while ensuring safety. You will be provided with a script in json format for a video containing information from visual captions and audio transcripts. Each entry in the script follows the format:

    {{
    "start_time":"start_time_in_seconds",
    "end_time": "end_time_in_seconds",
    "audio_transcript": "the_transcript_from_automatic_speech_recognition_system",
    "visual_caption": "the_caption_of_the_visuals_using_computer_vision_system"
    }}
    Note that the audio_transcript can sometimes be empty.

    Ensure you do not introduce any new named entities in your output and maintain the utmost factual accuracy in your responses.

    In the addition you will be provided with title of video extracted.
    """
    instruction = (
        "Provide a short and concise answer to the following user's question. "
        "Avoid mentioning any details about the script in JSON format. "
        "For example, a good response would be: 'Based on the analysis, "
        "here are the most relevant/useful/aesthetic moments.' "
        "A less effective response would be: "
        "'Based on the provided visual caption/audio transcript, "
        "here are the most relevant/useful/aesthetic moments. The user question is "
    )

    user_prompt_template = (
        "{instruction}"
        "Given the timeline of audio and visual activities in the video below "
        "I want to find out the following: {question}"
        "The timeline is: "
        "{timeline}"
        "\n"
        "The title of the video is {video_title}"
    )

    messages = []
    messages.append(ChatMessage(content=system_prompt_preamble, role="system"))
    messages.append(
        ChatMessage(
            content=user_prompt_template.format(
                instruction=instruction,
                question=question,
                timeline=timeline,
                video_title=metadata.title,
            ),
            role="user",
        )
    )

    dialog = ChatDialog(messages=messages)
    return dialog



import uuid
from enum import Enum

from sqlalchemy import (
    JSON,
    UUID,
    Column,
    DateTime,
    Float,
    Integer,
    PickleType,
    String,
)
from sqlalchemy import Enum as SqlEnum

from aana.storage.models.base import BaseEntity, TimeStampEntity


class Status(str, Enum):
    """Enum for task status."""

    CREATED = "created"
    ASSIGNED = "assigned"
    COMPLETED = "completed"
    RUNNING = "running"
    FAILED = "failed"
    NOT_FINISHED = "not_finished"


class TaskEntity(BaseEntity, TimeStampEntity):
    """Table for task items."""

    __tablename__ = "tasks"
    id = Column(UUID, primary_key=True, default=uuid.uuid4, comment="Task ID")
    endpoint = Column(
        String, nullable=False, comment="The endpoint to which the task is assigned"
    )
    data = Column(PickleType, nullable=False, comment="Data for the task")
    status = Column(
        SqlEnum(Status),
        nullable=False,
        default=Status.CREATED,
        comment="Status of the task",
    )
    priority = Column(
        Integer,
        nullable=False,
        default=0,
        comment="Priority of the task (0 is the lowest)",
    )
    assigned_at = Column(
        DateTime(timezone=True),
        nullable=True,
        comment="Timestamp when the task was assigned",
    )
    completed_at = Column(
        DateTime(timezone=True),
        nullable=True,
        comment="Timestamp when the task was completed",
    )
    progress = Column(
        Float, nullable=False, default=0.0, comment="Progress of the task in percentage"
    )
    result = Column(JSON, nullable=True, comment="Result of the task in JSON format")

    def __repr__(self):
        """String representation of the task."""
        return f"<TaskEntity(id={self.id}, endpoint={self.endpoint}, status={self.status}, priority={self.priority}, progress={self.progress})>"



# ruff: noqa: S101
from importlib import resources

import pytest
from sqlalchemy.orm import Session

from aana.core.models.asr import (
    AsrSegment,
    AsrSegments,
    AsrTranscription,
    AsrTranscriptionInfo,
)
from aana.core.models.captions import Caption, CaptionsList
from aana.core.models.time import TimeInterval
from aana.core.models.video import Video
from aana.storage.services.video import (
    save_transcripts_batch,
    save_video,
    save_video_batch,
    save_video_captions,
    save_video_transcription,
)


@pytest.fixture()
def mock_session(mocker):
    """Patches the Session object with a mock."""
    session_mock = mocker.MagicMock(spec=Session)
    context_var_mock = mocker.MagicMock(spec=Session)
    # Ensure that the object used inside a with block is the same.
    # Using `session_mock` doesn't work here, perhaps because it creates a
    # reference cycle.
    session_mock.return_value.__enter__.return_value = context_var_mock
    # Ensure that the context var is visible on the injected mock.
    session_mock.context_var = context_var_mock
    # Emulate the behavior of the empty database.
    context_var_mock.query.return_value.filter_by.return_value.first.return_value = None
    context_var_mock.query.return_value.get.return_value = None
    mocker.patch("aana.storage.services.video.Session", session_mock)
    return session_mock


def test_save_video(mock_session):
    """Tests save media function."""
    media_id = "foobar"
    duration = 550.25
    path = resources.path("aana.tests.files.videos", "squirrel.mp4")
    video = Video(path=path, media_id=media_id)
    result = save_video(video, duration)

    assert result["media_id"] == media_id
    # once each for MediaEntity and VideoEntity
    assert mock_session.context_var.add.call_count == 2
    assert mock_session.context_var.commit.call_count == 2


def test_save_videos_batch(mock_session):
    """Tests save media function."""
    media_ids = ["foo", "bar"]
    path = resources.path("aana.tests.files.videos", "squirrel.mp4")
    videos = [Video(path=path, media_id=m_id) for m_id in media_ids]
    durations = [0.1] * len(media_ids)

    result = save_video_batch(videos, durations)

    assert result["media_ids"] == media_ids
    # once each for MediaEntities and VideoEntities
    assert mock_session.context_var.add_all.call_count == 2
    assert mock_session.context_var.commit.call_count == 2


def test_save_video_transcription(mock_session):
    """Tests save transcripts function."""
    media_id = "0"
    model = "test_model"
    transcript = "This is a transcript. And this is another sentence. And a third one."
    lang = "en"
    lang_conf = 0.5
    transcript = AsrTranscription(text=transcript)
    transcription_info = AsrTranscriptionInfo(
        language=lang, language_confidence=lang_conf
    )

    segments = AsrSegments(
        [
            AsrSegment(
                text="This is a transcript.",
                time_interval=TimeInterval(start=0, end=1),
                confidence=0.99,
                no_speech_confidence=0.1,
            ),
            AsrSegment(
                text="And this is another sentence.",
                time_interval=TimeInterval(start=1, end=2),
                confidence=0.99,
                no_speech_confidence=0.1,
            ),
            AsrSegment(
                text="And a third one.",
                time_interval=TimeInterval(start=2, end=3),
                confidence=0.99,
                no_speech_confidence=0.1,
            ),
        ]
    )

    result = save_video_transcription(
        model_name=model,
        media_id=media_id,
        transcription_info=transcription_info,
        transcription=transcript,
        segments=segments,
    )
    assert "transcription_id" in result

    mock_session.context_var.add_all.assert_called_once()
    # Once for media, once for video, once for transcript
    assert mock_session.context_var.commit.call_count == 1


def test_save_transcripts_batch(mock_session):
    """Tests save transcripts function."""
    media_ids = ["0", "1", "2"]
    model = "test_model"
    texts = ("A transcript", "Another transcript", "A third transcript")
    infos = [("en", 0.5), ("de", 0.36), ("fr", 0.99)]
    transcripts = [AsrTranscription(text=text) for text in texts]
    transcription_infos = [
        AsrTranscriptionInfo(language=lang, language_confidence=conf)
        for lang, conf in infos
    ]
    segments = [
        [
            AsrSegment(
                text="",
                time_interval=TimeInterval(start=0, end=1),
                confidence=0.99,
                no_speech_confidence=0.1,
            )
        ]
        * 5
    ] * 3
    result = save_transcripts_batch(
        model, media_ids, transcription_infos, transcripts, segments
    )
    result_ids = result["transcription_ids"]

    assert (
        len(result_ids) == len(transcripts) == len(transcription_infos) == len(segments)
    )
    mock_session.context_var.add_all.assert_called_once()
    mock_session.context_var.commit.assert_called_once()


def test_save_captions_single(mock_session):
    """Tests save captions function."""
    media_id = "0"
    model_name = "test_model"
    captions = ["A caption", "Another caption", "A third caption"]
    captions_list = CaptionsList([Caption(caption) for caption in captions])
    timestamps = [0.1, 0.2, 0.3]
    frame_ids = [0, 1, 2]

    result = save_video_captions(
        model_name, media_id, captions_list, timestamps, frame_ids
    )

    assert (
        len(result["caption_ids"])
        == len(captions_list)
        == len(timestamps)
        == len(frame_ids)
    )
    mock_session.context_var.add_all.assert_called_once()
    # Once for media, once for video, once for captions
    assert mock_session.context_var.commit.call_count == 1



from enum import Enum
from pathlib import Path

import orjson
from alembic import command
from alembic.config import Config
from sqlalchemy import create_engine

from aana.utils.json import orjson_serializer


class DbType(str, Enum):
    """Engine types for relational database."""

    POSTGRESQL = "postgresql"
    SQLITE = "sqlite"


def create_postgresql_engine(config):
    """Create PostgreSQL engine based on the provided configuration.

    Args:
        config (PostgreSQLConfig): Configuration for PostgreSQL.

    Returns:
        sqlalchemy.engine.Engine: SQLAlchemy engine instance.
    """
    connection_string = f"postgresql://{config['user']}:{config['password']}@{config['host']}:{config['port']}/{config['database']}"
    return create_engine(
        connection_string,
        json_serializer=lambda obj: orjson_serializer(obj).decode(),
        json_deserializer=orjson.loads,
    )


def create_sqlite_engine(config):
    """Create an SQLite SQLAlchemy engine based on the provided configuration.

    Args:
        config (SQLite config): Configuration for SQLite.

    Returns:
        sqlalchemy.engine.Engine: SQLAlchemy engine instance.
    """
    connection_string = f"sqlite:///{config['path']}"
    return create_engine(
        connection_string,
        json_serializer=lambda obj: orjson_serializer(obj).decode(),
        json_deserializer=orjson.loads,
    )


def create_database_engine(db_config):
    """Create SQLAlchemy engine based on the provided configuration.

    Args:
        db_config (DbConfig): Database configuration.

    Returns:
        sqlalchemy.engine.Engine: SQLAlchemy engine instance.
    """
    db_type = getattr(db_config, "datastore_type", "").lower()

    if db_type == DbType.POSTGRESQL:
        return create_postgresql_engine(db_config.datastore_config)
    elif db_type == DbType.SQLITE:
        return create_sqlite_engine(db_config.datastore_config)
    else:
        raise ValueError(f"Unsupported database type: {db_type}")  # noqa: TRY003


def get_alembic_config(
    app_config, ini_file_path: Path, alembic_data_path: Path
) -> Config:
    """Produces an alembic config to run migrations programmatically."""
    engine = app_config.db_config.get_engine()
    alembic_config = Config(ini_file_path)
    alembic_config.set_main_option("script_location", str(alembic_data_path))
    config_section = alembic_config.get_section(alembic_config.config_ini_section, {})
    config_section["sqlalchemy.url"] = engine.url

    return alembic_config


def run_alembic_migrations(settings):
    """Runs alembic migrations before starting up."""
    # We need the path to aana/alembic and aana/alembic.ini
    # This is a hack until we need something better.
    current_path = Path(__file__)
    aana_root = current_path.parent.parent  # go up two directories
    if aana_root.name != "aana":  # we are not in the right place
        raise RuntimeError("Not in right directory, exiting.")  # noqa: TRY003
    ini_file_path = aana_root / "alembic.ini"
    alembic_data_path = aana_root / "alembic"

    alembic_config = get_alembic_config(settings, ini_file_path, alembic_data_path)
    engine = settings.db_config.get_engine()
    with engine.begin() as connection:
        alembic_config.attributes["connection"] = connection
        command.upgrade(alembic_config, "head")


def drop_all_tables(settings):
    """Drops all tables in the database."""
    # TODO: only allow this in testing mode
    current_path = Path(__file__)
    aana_root = current_path.parent.parent  # go up two directories
    if aana_root.name != "aana":  # we are not in the right place
        raise RuntimeError("Not in right directory, exiting.")  # noqa: TRY003
    ini_file_path = aana_root / "alembic.ini"
    alembic_data_path = aana_root / "alembic"

    alembic_config = get_alembic_config(settings, ini_file_path, alembic_data_path)
    command.downgrade(alembic_config, "base")



from collections import defaultdict
from math import floor

from aana.core.models.asr import AsrSegments
from aana.core.models.audio import Audio
from aana.core.models.video import Video
from aana.integrations.external.av import load_audio


def extract_audio(video: Video) -> Audio:
    """Extract the audio file from a Video and return an Audio object.

    Args:
        video (Video): The video file to extract audio.

    Returns:
        Audio: an Audio object containing the extracted audio.
    """
    audio_bytes = load_audio(video.path)

    # Only difference will be in path where WAV file will be stored
    # and in content but has same media_id
    return Audio(
        url=video.url,
        media_id=f"audio_{video.media_id}",
        content=audio_bytes,
        title=video.title,
        description=video.description,
    )


def generate_combined_timeline(
    transcription_segments: AsrSegments,
    captions: list[str],
    caption_timestamps: list[float],
    chunk_size: float = 10.0,
):
    """Generates a combined timeline from the ASR segments and the captions.

    Args:
        transcription_segments (AsrSegments): the ASR segments
        captions (list[str]): the captions
        caption_timestamps (list[float]): the timestamps for the captions
        chunk_size (float, optional): the chunk size for the combined timeline in seconds. Defaults to 10.0.

    Returns:
        dict: dictionary containing one key, "timeline", which is a list of dictionaries with the following keys:
            "start_time": the start time of the chunk in seconds
            "end_time": the end time of the chunk in seconds
            "audio_transcript": the audio transcript for the chunk
            "visual_caption": the visual caption for the chunk
    """
    timeline_dict: defaultdict[int, dict[str, list[str]]] = defaultdict(
        lambda: {"transcription": [], "captions": []}
    )
    for segment in transcription_segments:
        segment_start = segment.time_interval.start
        chunk_index = floor(segment_start / chunk_size)
        timeline_dict[chunk_index]["transcription"].append(segment.text)

    if len(captions) != len(caption_timestamps):
        raise ValueError(  # noqa: TRY003
            f"Length of captions ({len(captions)}) and timestamps ({len(caption_timestamps)}) do not match"
        )

    for timestamp, caption in zip(caption_timestamps, captions, strict=True):
        chunk_index = floor(timestamp / chunk_size)
        timeline_dict[chunk_index]["captions"].append(caption)

    num_chunks = max(timeline_dict.keys()) + 1

    timeline = [
        {
            "start_time": chunk_index * chunk_size,
            "end_time": (chunk_index + 1) * chunk_size,
            "audio_transcript": "\n".join(timeline_dict[chunk_index]["transcription"]),
            "visual_caption": "\n".join(timeline_dict[chunk_index]["captions"]),
        }
        for chunk_index in range(num_chunks)
    ]

    return {
        "timeline": timeline,
    }



from sqlalchemy import Column, DateTime, MetaData, func
from sqlalchemy.orm import DeclarativeBase


class BaseEntity(DeclarativeBase):
    """Base for all ORM classes."""

    metadata = MetaData(
        naming_convention={
            "ix": "ix_%(column_0_label)s",
            "uq": "uq_%(table_name)s_%(column_0_name)s",
            "ck": "ck_%(table_name)s_`%(constraint_name)s`",
            "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
            "pk": "pk_%(table_name)s",
        }
    )


class TimeStampEntity:
    """Mixin for database entities that will have create/update timestamps."""

    create_ts = Column(
        DateTime(timezone=True),
        server_default=func.now(),
        comment="Timestamp when row is inserted",
    )
    update_ts = Column(
        DateTime(timezone=True),
        onupdate=func.now(),
        comment="Timestamp when row is updated",
    )



# ruff: noqa: S101
import pytest
from ray import serve

from aana.core.models.chat import ChatDialog, ChatMessage
from aana.core.models.sampling import SamplingParams
from aana.exceptions.runtime import PromptTooLongException
from aana.tests.utils import (
    compare_texts,
    get_deployments_by_type,
    is_gpu_available,
    is_using_deployment_cache,
)


def get_expected_output(name):
    """Gets expected output for a given text_generation model."""
    if name == "vllm_llama2_7b_chat_deployment":
        return (
            "  Elon Musk is a South African-born entrepreneur, inventor, "
            "and business magnate who is best known for his innovative companies in"
        )
    elif name == "meta_llama3_8b_instruct_deployment":
        return (
            " Elon Musk is a South African-born entrepreneur, inventor,"
            "and business magnate. He is the CEO and CTO of SpaceX, "
            "CEO and product architect of Tesla"
        )
    elif name == "microsoft_phi_3_mini_instruct_deployment":
        return " Elon Musk is a business magnate, industrial designer, and engineer. He is the founder, CEO, CTO, and chief designer of Space"
    elif name == "hf_phi3_mini_4k_instruct_text_gen_deployment":
        return "Elon Musk is a prominent entrepreneur and business magnate known for his significant contributions to the technology and automotive industries. He was born"
    elif name == "internlm2_5_7b_chat_deployment":
        return "\nElon Musk is a prominent entrepreneur, inventor, and business magnate known for his contributions to a variety of high-tech industries. He is the founder"
    else:
        raise ValueError(f"Unknown deployment name: {name}")  # noqa: TRY003


def get_expected_chat_output(name):
    """Gets expected output for a given text_generation model."""
    if name == "vllm_llama2_7b_chat_deployment":
        return (
            "  Elon Musk is a South African-born entrepreneur, inventor, "
            "and business magnate who is best known for his innovative companies in"
        )
    elif name == "meta_llama3_8b_instruct_deployment":
        return "Elon Musk is a South African-born entrepreneur, inventor, and business magnate. He is best known for his ambitious goals to revolutionize the transportation, energy"
    elif name == "microsoft_phi_3_mini_instruct_deployment":
        return " Elon Musk is a business magnate, industrial designer, and engineer. He is the founder, CEO, CTO, and chief designer of Space"
    elif name == "hf_phi3_mini_4k_instruct_text_gen_deployment":
        return "Elon Musk is a prominent entrepreneur and business magnate known for his significant contributions to the technology and automotive industries. He was born"
    elif name == "internlm2_5_7b_chat_deployment":
        return "Elon Musk is a prominent entrepreneur, inventor, and business magnate known for his contributions to the fields of technology, space exploration, and sustainable energy."
    else:
        raise ValueError(f"Unknown deployment name: {name}")  # noqa: TRY003


def get_prompt(name):
    """Gets the prompt for a given text_generation model."""
    if name == "vllm_llama2_7b_chat_deployment":
        return "[INST] Who is Elon Musk? [/INST]"
    elif name == "hf_phi3_mini_4k_instruct_text_gen_deployment":
        return "<|user|>\ Who is Elon Musk? <|end|>\n<|assistant|>"
    elif name == "meta_llama3_8b_instruct_deployment":
        return "[INST] Who is Elon Musk? [/INST]"
    elif name == "microsoft_phi_3_mini_instruct_deployment":
        return "[INST] Who is Elon Musk? [/INST]"
    elif name == "internlm2_5_7b_chat_deployment":
        return "<s><|im_start|>user\nWho is Elon Musk?<|im_end|>\n<|im_start|>assistant"
    else:
        raise ValueError(f"Unknown deployment name: {name}")  # noqa: TRY003


long_context_models = ["internlm2_5_7b_chat_deployment"]


@pytest.fixture(
    scope="function",
    params=get_deployments_by_type("VLLMDeployment")
    + get_deployments_by_type("HfTextGenerationDeployment"),
)
def setup_text_generation_deployment(app_setup, request):
    """Setup text_generation deployment."""
    name, deployment = request.param
    deployments = [
        {
            "name": "text_generation_deployment",
            "instance": deployment,
        }
    ]
    endpoints = []

    return name, deployment, app_setup(deployments, endpoints)


@pytest.mark.skipif(
    not is_gpu_available() and not is_using_deployment_cache(),
    reason="GPU is not available",
)
@pytest.mark.asyncio
async def test_text_generation_deployments(setup_text_generation_deployment):
    """Test text_generation deployments."""
    name, deployment, app = setup_text_generation_deployment

    handle = serve.get_app_handle("text_generation_deployment")

    expected_text = get_expected_output(name)
    prompt = get_prompt(name)

    # test generate method
    output = await handle.generate.remote(
        prompt=prompt,
        sampling_params=SamplingParams(temperature=0.0, max_tokens=32),
    )
    text = output["text"]

    compare_texts(expected_text, text)

    # test generate_stream method
    stream = handle.options(stream=True).generate_stream.remote(
        prompt=prompt,
        sampling_params=SamplingParams(temperature=0.0, max_tokens=32),
    )
    text = ""
    async for chunk in stream:
        text += chunk["text"]

    compare_texts(expected_text, text)

    # test generate_batch method
    output = await handle.generate_batch.remote(
        prompts=[prompt, prompt],
        sampling_params=SamplingParams(temperature=0.0, max_tokens=32),
    )
    texts = output["texts"]

    for text in texts:
        compare_texts(expected_text, text)

    # test chat method
    expected_text = get_expected_chat_output(name)
    dialog = ChatDialog(
        messages=[
            ChatMessage(
                role="user",
                content="Who is Elon Musk?",
            )
        ]
    )
    output = await handle.chat.remote(
        dialog=dialog,
        sampling_params=SamplingParams(temperature=0.0, max_tokens=32),
    )

    response_message = output["message"]
    assert response_message.role == "assistant"
    text = response_message.content

    compare_texts(expected_text, text)

    # test chat_stream method
    stream = handle.options(stream=True).chat_stream.remote(
        dialog=dialog,
        sampling_params=SamplingParams(temperature=0.0, max_tokens=32),
    )

    text = ""
    async for chunk in stream:
        text += chunk["text"]

    compare_texts(expected_text, text)

    if name not in long_context_models:
        # test generate method with too long prompt
        with pytest.raises(PromptTooLongException):
            output = await handle.generate.remote(
                prompt=prompt * 1000,
                sampling_params=SamplingParams(temperature=0.0, max_tokens=32),
            )



from enum import Enum
from uuid import uuid4

from sqlalchemy import Column, String
from sqlalchemy.orm import relationship

from aana.storage.models.base import BaseEntity, TimeStampEntity
from aana.storage.types import MediaIdSqlType


class MediaType(str, Enum):
    """Enum for types of media file."""

    VIDEO = "video"


class MediaEntity(BaseEntity, TimeStampEntity):
    """Table for media items."""

    __tablename__ = "media"
    id = Column(
        MediaIdSqlType,
        primary_key=True,
        default=lambda: str(uuid4()),
        comment="Unique identifier for the media",
    )
    media_type = Column(String, comment="The type of media")

    video = relationship(
        "VideoEntity",
        back_populates="media",
        cascade="all, delete",
        uselist=False,
    )



from collections.abc import AsyncGenerator
from typing import TYPE_CHECKING, Annotated, TypedDict

from pydantic import Field

from aana.api.api_generation import Endpoint
from aana.core.models.asr import (
    AsrSegments,
    AsrTranscription,
    AsrTranscriptionInfo,
)
from aana.core.models.media import MediaId
from aana.core.models.vad import VadParams
from aana.core.models.video import VideoInput
from aana.core.models.whisper import BatchedWhisperParams, WhisperParams
from aana.deployments.aana_deployment_handle import AanaDeploymentHandle
from aana.integrations.external.yt_dlp import download_video
from aana.processors.remote import run_remote
from aana.processors.video import extract_audio
from aana.projects.whisper.const import asr_model_name
from aana.storage.services.video import (
    delete_media,
    load_video_transcription,
    save_video,
    save_video_transcription,
)

if TYPE_CHECKING:
    from aana.core.models.audio import Audio
    from aana.core.models.video import Video


class TranscribeVideoOutput(TypedDict):
    """The output of the transcribe video endpoint."""

    transcription: AsrTranscription
    transcription_info: AsrTranscriptionInfo
    segments: AsrSegments
    transcription_id: Annotated[int, Field(..., description="Transcription Id")]


class TranscriptionOutput(TypedDict):
    """The transcription output."""

    transcription: AsrTranscription
    segments: AsrSegments
    transcription_info: AsrTranscriptionInfo


class DeleteMediaOutput(TypedDict):
    """The output of the delete media endpoint."""

    media_id: MediaId


class SimpleTranscribeVideoOutput(TypedDict):
    """The output of the simple transcribe video endpoint."""

    transcription: AsrTranscription
    transcription_info: AsrTranscriptionInfo
    segments: AsrSegments


class SimpleTranscribeVideoEndpoint(Endpoint):
    """Simple video transcription (no saving)."""

    async def initialize(self):
        """Initialize the endpoint."""
        self.asr_handle = await AanaDeploymentHandle.create("asr_deployment")

    async def run(
        self, video: VideoInput, whisper_params: WhisperParams
    ) -> AsyncGenerator[SimpleTranscribeVideoOutput, None]:
        """Transcribe video."""
        video_obj: Video = await run_remote(download_video)(video_input=video)
        audio: Audio = extract_audio(video=video_obj)

        transcription_list = []
        segments_list = []
        transcription_info_list = []
        async for whisper_output in self.asr_handle.transcribe_stream(
            audio=audio, params=whisper_params
        ):
            transcription_list.append(whisper_output["transcription"])
            segments_list.append(whisper_output["segments"])
            transcription_info_list.append(whisper_output["transcription_info"])
            yield {
                "transcription": whisper_output["transcription"],
                "segments": whisper_output["segments"],
                "info": whisper_output["transcription_info"],
            }


class TranscribeVideoEndpoint(Endpoint):
    """Transcribe video endpoint."""

    async def initialize(self):
        """Initialize the endpoint."""
        self.asr_handle = await AanaDeploymentHandle.create("asr_deployment")
        await super().initialize()

    async def run(
        self, video: VideoInput, whisper_params: WhisperParams
    ) -> AsyncGenerator[TranscribeVideoOutput, None]:
        """Transcribe video."""
        video_obj: Video = await run_remote(download_video)(video_input=video)
        audio: Audio = extract_audio(video=video_obj)

        transcription_list = []
        segments_list = []
        transcription_info_list = []
        async for whisper_output in self.asr_handle.transcribe_stream(
            audio=audio, params=whisper_params
        ):
            transcription_list.append(whisper_output["transcription"])
            segments_list.append(whisper_output["segments"])
            transcription_info_list.append(whisper_output["transcription_info"])
            yield {
                "transcription": whisper_output["transcription"],
                "segments": whisper_output["segments"],
                "info": whisper_output["transcription_info"],
            }
        transcription = sum(transcription_list, AsrTranscription())
        segments = sum(segments_list, AsrSegments())
        transcription_info = sum(transcription_info_list, AsrTranscriptionInfo())

        save_video(
            video=video_obj, duration=0.0
        )  # set duration to 0.0 as we don't have the actual duration

        transcription_record = save_video_transcription(
            model_name=asr_model_name,
            media_id=video_obj.media_id,
            transcription=transcription,
            segments=segments,
            transcription_info=transcription_info,
        )
        yield {"transcription_id": transcription_record["transcription_id"]}


# TODO: Update once batched whisper PR is merged
# class TranscribeVideoInChunksEndpoint(Endpoint):
#     """Transcribe video in chunks endpoint."""

#     async def initialize(self):
#         """Initialize the endpoint."""
#         self.asr_handle = await AanaDeploymentHandle.create("asr_deployment")
#         self.vad_handle = await AanaDeploymentHandle.create("vad_deployment")
#         await super().initialize()

#     async def run(
#         self,
#         video: VideoInput,
#         whisper_params: BatchedWhisperParams,
#         vad_params: VadParams,
#     ) -> AsyncGenerator[TranscribeVideoOutput, None]:
#         """Transcribe video in chunks."""
#         video_obj: Video = await run_remote(download_video)(video_input=video)
#         audio: Audio = extract_audio(video=video_obj)

#         vad_output = await self.vad_handle.asr_preprocess_vad(
#             audio=audio, params=vad_params
#         )
#         vad_segments = vad_output["segments"]

#         transcription_list = []
#         segments_list = []
#         transcription_info_list = []
#         async for whisper_output in self.asr_handle.transcribe_in_chunks(
#             audio=audio, params=whisper_params, segments=vad_segments
#         ):
#             transcription_list.append(whisper_output["transcription"])
#             segments_list.append(whisper_output["segments"])
#             transcription_info_list.append(whisper_output["transcription_info"])
#             yield {
#                 "transcription": whisper_output["transcription"],
#                 "segments": whisper_output["segments"],
#                 "info": whisper_output["transcription_info"],
#             }
#         transcription = sum(transcription_list, AsrTranscription())
#         segments = sum(segments_list, AsrSegments())
#         transcription_info = sum(transcription_info_list, AsrTranscriptionInfo())

#         save_video(
#             video=video_obj, duration=0.0
#         )  # set duration to 0.0 as we don't have the actual duration

#         transcription_record = save_video_transcription(
#             model_name=asr_model_name,
#             media_id=video_obj.media_id,
#             transcription=transcription,
#             segments=segments,
#             transcription_info=transcription_info,
#         )
#         yield {"transcription_id": transcription_record["transcription_id"]}


class LoadTranscriptionEndpoint(Endpoint):
    """Load transcription endpoint."""

    async def run(self, media_id: MediaId) -> TranscriptionOutput:
        """Load transcription."""
        transcription_record = load_video_transcription(
            model_name=asr_model_name, media_id=media_id
        )
        return {
            "transcription": transcription_record["transcription"],
            "segments": transcription_record["segments"],
            "info": transcription_record["transcription_info"],
        }


class DeleteMediaEndpoint(Endpoint):
    """Delete media endpoint."""

    async def run(self, media_id: MediaId) -> DeleteMediaOutput:
        """Delete media."""
        delete_media(media_id=media_id)
        return {"media_id": media_id}



import asyncio
import threading
from typing import Any


def run_async(coro: asyncio.coroutine) -> Any:
    """Run a coroutine in a thread if the current thread is running an event loop.

    Otherwise, run the coroutine in the current asyncio loop.

    Useful when you want to run an async function in a non-async context.

    From: https://stackoverflow.com/a/75094151

    Args:
        coro (Coroutine): The coroutine to run.

    Returns:
        Any: The result of the coroutine.
    """

    class RunThread(threading.Thread):
        """Run a coroutine in a thread."""

        def __init__(self, coro):
            """Initialize the thread."""
            self.coro = coro
            self.result = None
            super().__init__()

        def run(self):
            """Run the coroutine."""
            self.result = asyncio.run(self.coro)

    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        loop = None
    if loop and loop.is_running():
        thread = RunThread(coro)
        thread.start()
        thread.join()
        return thread.result
    else:
        return asyncio.run(coro)



# ruff: noqa: S101

import pytest
from ray import serve

from aana.core.models.chat import Prompt
from aana.tests.utils import (
    get_deployments_by_type,
    is_gpu_available,
    is_using_deployment_cache,
)


@pytest.fixture(
    scope="function", params=get_deployments_by_type("StableDiffusion2Deployment")
)
def setup_deployment(app_setup, request):
    """Setup Stable Diffusion 2 deployment."""
    name, deployment = request.param
    deployments = [
        {
            "name": "sd2_deployment",
            "instance": deployment,
        }
    ]
    endpoints = []

    return name, deployment, app_setup(deployments, endpoints)


@pytest.mark.skipif(
    not is_gpu_available() and not is_using_deployment_cache(),
    reason="GPU is not available",
)
@pytest.mark.asyncio
@pytest.mark.parametrize(
    "prompt",
    ["Mona Lisa but from Picasso's Blue Period"],
)
async def test_stablediffusion2_deployment(setup_deployment, prompt):
    """Test HuggingFace BLIP2 deployments."""
    handle = serve.get_app_handle("sd2_deployment")

    output = await handle.generate.remote(prompt=Prompt(prompt))

    image = output["image"]

    assert image is not None
    assert image.size == (768, 768)
    assert image.mode == "RGB"



# ruff: noqa: S101
# Test llama2 endpoints.


import pytest

from aana.tests.utils import is_gpu_available, is_using_deployment_cache

TARGET = "llama2"

LLM_GENERATE = "/llm/generate"
LLM_GENERATE_STREAM = "/llm/generate_stream"
LLM_CHAT = "/llm/chat"
LLM_CHAT_STREAM = "/llm/chat_stream"


@pytest.mark.skipif(
    not is_gpu_available() and not is_using_deployment_cache(),
    reason="GPU is not available",
)
@pytest.mark.parametrize("call_endpoint", [TARGET], indirect=True)
@pytest.mark.parametrize(
    "prompt, error",
    [
        ("[INST] Who is Elon Musk? [/INST]", None),
        ("[INST] Where is the Eiffel Tower? [/INST]", None),
        ("[INST] Who is Elon Musk? [/INST]" * 1000, "PromptTooLongException"),
    ],
)
def test_llama_generate(call_endpoint, prompt, error):
    """Test llama generate endpoints."""
    call_endpoint(
        LLM_GENERATE,
        {"prompt": prompt},
        expected_error=error,
    )

    call_endpoint(
        LLM_GENERATE_STREAM,
        {"prompt": prompt},
        expected_error=error,
    )


@pytest.mark.skipif(
    not is_gpu_available() and not is_using_deployment_cache(),
    reason="GPU is not available",
)
@pytest.mark.parametrize("call_endpoint", [TARGET], indirect=True)
@pytest.mark.parametrize(
    "dialog, error",
    [
        ({"messages": [{"role": "user", "content": "Who is Elon Musk?"}]}, None),
        (
            {"messages": [{"role": "user", "content": "Where is the Eiffel Tower?"}]},
            None,
        ),
        (
            {"messages": [{"role": "user", "content": "Who is Elon Musk?" * 100}]},
            None,
        ),
        (
            {"messages": [{"role": "user", "content": "Who is Elon Musk?" * 1000}]},
            "PromptTooLongException",
        ),
    ],
)
def test_llama_chat(call_endpoint, dialog, error):
    """Test llama chat endpoint."""
    call_endpoint(
        LLM_CHAT,
        {"dialog": dialog},
        expected_error=error,
    )

    call_endpoint(
        LLM_CHAT_STREAM,
        {"dialog": dialog},
        expected_error=error,
    )
